
from __future__ import annotations
from typing import TYPE_CHECKING, Any, TypeVar, Optional, Mapping, Union, Callable, overload
if TYPE_CHECKING:
    from types import TracebackType
    from .auth.types import AuthorizationGrant
    from .http.types import RequestFiles
    from .http.payload import Payload
    from .core.http_client_SYNC import HTTPClient
    from .types import JSON_ro

from .auth import Token
from .auth import grants
from .core import grants as core_grants
from .core.http_client_SYNC import (
    RedditHTTPClient,
    build_reddit_http_client,
    build_reddit_http_client_from_access_token,
)
from .exceptions import raise_for_reddit_error, raise_for_non_json_response
from .http.util.json_loading import load_json_from_response
from .util.redditwarp_installed_client_credentials import get_redditwarp_client_id, get_device_id


class Client:
    """Gateway to interacting with the Reddit API."""

    _TSelf = TypeVar('_TSelf', bound='Client')

    @staticmethod
    def from_praw_config(section_name: str, *, filepath: Optional[str] = None) -> Client:
        """Initialize a `Client` instance from a `praw.ini` file.

        This method aims to replicate the single-argument form of PRAW's `Reddit`
        class constructor. If `filepath` is not specified it will search for
        `praw.ini` files in the same locations PRAW does.

        Only a subset of PRAW's configuration keys are read:

        * `client_id`
        * `client_secret`
        * `refresh_token`
        * `username`
        * `password`
        * `user_agent`

        The credential values are given directly to the `Client` constructor,
        then the `user_agent` value (if present) is passed to :meth:`.set_user_agent`.

        .. .PARAMETERS

        :param `str` section_name:
            The section name of the ini file in which to read values from.
            Pass an empty string to use the default section name "`DEFAULT`".
        :param `Optional[str]` filepath:
            The location of a `praw.ini` file to read.

            If not specified, the locations returned by
            :func:`redditwarp.util.praw_config.get_praw_ini_potential_locations`
            are searched and any files found are read and combined into a single
            configuration.
        """
        from .util.praw_config_SYNC import client_from_praw_config  # Avoid cyclic import
        return client_from_praw_config(section_name, filepath=filepath)

    @classmethod
    def from_http(cls: type[_TSelf], http: HTTPClient) -> _TSelf:
        """Alternative constructor for testing purposes or advanced use cases."""
        self = cls.__new__(cls)
        self._init(http)
        return self

    @classmethod
    def from_access_token(cls: type[_TSelf], access_token: str) -> _TSelf:
        """Construct an instance without a token client.

        No token client means `self.http.authorizer.token_client` will be `None`.

        When the access token becomes invalid you'll need to deal with the
        401 Unauthorized :class:`~redditwarp.http.exceptions.StatusCodeException`
        exception that will be thrown upon making API requests.

        Use the :meth:`.set_access_token` method to assign new access tokens.
        """
        http = build_reddit_http_client_from_access_token(access_token)
        return cls.from_http(http)

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, /) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, /, *, grant: AuthorizationGrant) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, refresh_token: str, /) -> None: ...
    @overload
    def __init__(self, client_id: str, client_secret: str, username: str, password: str, /) -> None: ...
    def __init__(self, *creds: str, grant: Optional[AuthorizationGrant] = None) -> None:
        """
        .. .PARAMETERS

        :param `str` client_id:
        :param `str` client_secret:
        :param `str` refresh_token:
        :param `str` username:
        :param `str` password:
        :param grant:
            Specify an explicit grant. Use this parameter if you want to limit
            authorization scopes, or if you need to use the Installed Client grant type.
        :type grant: :obj:`~.auth.types.AuthorizationGrant`

        If `client_id` and `client_secret` are the only credentials given then a
        Client Credentials grant will be configured. The client will effectively
        be in a read-only mode.
        """
        client_id = client_secret = ''
        n = len(creds)
        if n == 0:
            client_id = get_redditwarp_client_id()
            grant = core_grants.InstalledClientGrant(get_device_id())
        elif n == 2:
            client_id, client_secret = creds
            if grant is None:
                grant = grants.ClientCredentialsGrant()
        elif n == 3:
            client_id, client_secret, refresh_token = creds
            grant = grants.RefreshTokenGrant(refresh_token)
        elif n == 4:
            client_id, client_secret, username, password = creds
            grant = grants.ResourceOwnerPasswordCredentialsGrant(username, password)
        else:
            raise TypeError

        http = build_reddit_http_client(client_id, client_secret, grant)
        self._init(http)

    def _init(self, http: HTTPClient) -> None:
        self.http: HTTPClient = http
        self.last_value: Any = None

        # Delay heavy import till client instantiation
        # instead of library import.
        from .siteprocs.SYNC import SiteProcedures
        self.p: SiteProcedures = SiteProcedures(self)

    def __enter__(self: _TSelf) -> _TSelf:
        return self

    def __exit__(self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        """Calls `self.close()`"""
        self.close()
        return None

    def close(self) -> None:
        """Calls `self.http.close()`"""
        self.http.close()

    def request(self,
        verb: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: JSON_ro = None,
        files: Optional[RequestFiles] = None,
        payload: Optional[Payload] = None,
        timeout: float = -2,
        follow_redirects: Optional[bool] = None,
        snub: Optional[Callable[[JSON_ro], None]] = raise_for_reddit_error,
    ) -> Any:
        """Make an API request and return JSON data.

        The parameters are similar to
        :meth:`HTTPClient.request <redditwarp.http.http_client_SYNC.HTTPClient.request>`,
        except for `snub`.

        The `snub` function examines the returned JSON data for API problems and
        generates exceptions based on them. You may choose to assign this option
        if you implement an API endpoint and know the structure of the errors,
        but the default snub function covers most Reddit API error structures.

        This method is only appropriate for making calls to the Reddit API
        and not any other website because of the domain specific post processing
        that happens with the response data.

        Below is a list of the main exception types thrown by this method, ordered
        by precedence if multiple exceptions apply:

        - `redditwarp.exceptions.RedditError`
        - `redditwarp.http.exceptions.StatusCodeException`
        - `ValueError`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            An API error was detected. Thrown by the `snub` function.
        :raises redditwarp.http.exceptions.StatusCodeException:
            The request returned a non 200 status.
        :raises ValueError:
            The endpoint did not return JSON.
        """
        json_data = None
        try:
            resp = self.http.request(verb, url, params=params, headers=headers,
                    data=data, json=json, files=files, payload=payload,
                    timeout=timeout, follow_redirects=follow_redirects)

            if resp.data:
                try:
                    json_data = load_json_from_response(resp)
                except ValueError as cause:
                    try:
                        raise_for_non_json_response(resp)
                    except Exception as exc:
                        raise exc from cause
                    raise

                if snub is not None:
                    snub(json_data)

            resp.ensure_successful_status()
        finally:
            self.last_value = json_data
        return json_data

    def set_access_token(self, access_token: str) -> None:
        """Manually set the current access token."""
        http = self.http
        if not isinstance(http, RedditHTTPClient):
            raise RuntimeError(f"self.http must be {RedditHTTPClient.__name__}")
        http.authorizer.set_token(Token(access_token))

    def set_user_agent(self, s: Optional[str]) -> None:
        """Set a user agent description.

        To view or set the current user agent string directly,
        use `self.http.get_user_agent()`.
        """
        ua = self.http.user_agent_base
        if s is not None:
            ua = f"{ua} Bot !-- {s}"
        self.http.set_user_agent(ua)

RedditClient = Client
Reddit = Client
RedditWarp = Client



"""Extract the ID from a Reddit submission or comment URL."""

import urllib.parse
from pathlib import Path
from collections import deque

def extract_submission_id_from_url(url: str) -> int:
    urlparts = urllib.parse.urlsplit(url)
    if urlparts.scheme not in {'https', 'http'}:
        raise ValueError
    if urlparts.netloc not in {'www.reddit.com', 'reddit.com', 'redd.it', 'ssl.reddit.com'}:
        raise ValueError

    parts = deque(Path(urlparts.path).parts)

    def popq() -> str:
        try:
            return parts.popleft()
        except IndexError as cause:
            raise ValueError from cause

    t = popq()
    assert t == '/'
    val = ''
    t = popq()
    if t == 'comments':
        val = popq()
        if len(parts) > 1:
            raise ValueError
    elif t == 'r':
        popq()  # subreddit
        t = popq()
        if t == 'comments':
            val = popq()
            if len(parts) > 1:
                raise ValueError
        else:
            val = t
            if parts:
                raise ValueError
    else:
        val = t
        if parts:
            raise ValueError

    return int(val, 36)

def extract_comment_id_from_url(url: str) -> int:
    urlparts = urllib.parse.urlsplit(url)
    if urlparts.scheme not in {'https', 'http'}:
        raise ValueError
    if urlparts.netloc not in {'www.reddit.com', 'reddit.com', 'redd.it', 'ssl.reddit.com'}:
        raise ValueError

    parts = deque(Path(urlparts.path).parts)

    def popq() -> str:
        try:
            return parts.popleft()
        except IndexError as cause:
            raise ValueError from cause

    t = popq()
    assert t == '/'
    val = ''
    t = popq()
    if t == 'comments':
        popq()  # submission ID36
        popq()  # slug
        val = popq()
    elif t == 'r':
        popq()  # subreddit
        comments = popq()
        if comments != 'comments':
            raise ValueError
        popq()  # submission ID36
        popq()  # slug
        val = popq()

    return int(val, 36)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_ASYNC import Comment
    from .ASYNC import CommentProcedures

from ...util.base_conversion import to_base36
from ...util.extract_id_from_url import extract_comment_id_from_url
from ...exceptions import NoResultException

class Fetch:
    def __init__(self, outer: CommentProcedures, client: Client) -> None:
        self._outer = outer
        self._client = client

    async def __call__(self, idn: int) -> Comment:
        id36 = to_base36(idn)
        return await self.by_id36(id36)

    async def by_id36(self, id36: str) -> Comment:
        v = await self._outer.get.by_id36(id36)
        if v is None:
            raise NoResultException('target not found')
        return v

    async def by_url(self, url: str) -> Comment:
        return await self(extract_comment_id_from_url(url))




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping, TypeVar, Generic
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from .SYNC import SubmissionProcedures

from ...model_loaders.submission_SYNC import load_submission
from ...models.submission_SYNC import Submission, LinkPost, TextPost
from ...util.base_conversion import to_base36
from ...util.extract_id_from_url import extract_submission_id_from_url
from ...exceptions import NoResultException, RejectedResultException

T = TypeVar('T')

class Common(Generic[T]):
    def __init__(self, client: Client) -> None:
        self._client = client

    def __call__(self, idn: int) -> T:
        id36 = to_base36(idn)
        return self.by_id36(id36)

    def _load_object(self, m: Mapping[str, Any]) -> T:
        raise NotImplementedError

    def by_id36(self, id36: str) -> T:
        full_id36 = 't3_' + id36
        root = self._client.request('GET', '/api/info', params={'id': full_id36})
        if children := root['data']['children']:
            return self._load_object(children[0]['data'])
        raise NoResultException('target not found')

    def by_url(self, url: str) -> T:
        return self(extract_submission_id_from_url(url))



class Fetch(Common[Submission]):
    class AsTextPost(Common[TextPost]):
        def _load_object(self, m: Mapping[str, Any]) -> TextPost:
            post = load_submission(m, self._client)
            if isinstance(post, TextPost):
                return post
            raise RejectedResultException('the submission is not a text post')

    class AsLinkPost(Common[LinkPost]):
        def _load_object(self, m: Mapping[str, Any]) -> LinkPost:
            post = load_submission(m, self._client)
            if isinstance(post, LinkPost):
                return post
            raise RejectedResultException('the submission is not a link post')

    def __init__(self, outer: SubmissionProcedures, client: Client) -> None:
        super().__init__(client)
        self.as_textpost: Fetch.AsTextPost = self.AsTextPost(client)
        self.as_linkpost: Fetch.AsLinkPost = self.AsLinkPost(client)

    def _load_object(self, m: Mapping[str, Any]) -> Submission:
        return load_submission(m, self._client)




from __future__ import annotations
from typing import Mapping, Any

from datetime import datetime, timezone

from .datamemento import DatamementoPropertiesMixin

class User(DatamementoPropertiesMixin):
    class Subreddit:
        def __init__(self, d: Mapping[str, Any]) -> None:
            d = d['subreddit']
            self.name: str = d['display_name']
            ("")
            self.openness: str = d['subreddit_type']
            ("")
            self.subscriber_count: int = d['subscribers']
            ("")
            self.title: str = d['title']
            ("")
            self.public_description: str = d['public_description']
            ("")
            self.nsfw: bool = d['over_18']
            ("")

    class Me:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.friend: bool = d['is_friend']
            ("""
                Whether the user is marked as friend of the current user.

                If there is no user context, always false.

                This is an old reddit feature. See `<https://www.reddit.com/prefs/friends/>`_.
                """)
            self.blocked: bool = d['is_blocked']
            ("""
                True if the current account has blocked this user.

                If there is no user context, always false.
                """)

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id36: str = d['id']
        ("""
            User ID as a base 36 number.
            """)
        self.id: int = int(self.id36, 36)
        ("""
            User ID.
            """)
        self.name: str = d['name']
        ("""
            Username.
            """)
        self.created_ut: int = int(d['created_utc'])
        ("""
            Unix timestamp of when the user account was created.
            """)
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("""
            Datetime object of when the user account was created.
            """)

        self.post_karma: int = d['link_karma']
        ("")
        self.comment_karma: int = d['comment_karma']
        ("")
        # The user objects from wiki related data don't have these fields
        # but it uses a different user class type for now anyways.
        self.awardee_karma: int = d.get('awardee_karma', 0)
        ("")
        self.awarder_karma: int = d.get('awarder_karma', 0)
        ("")
        self.total_karma: int = d.get('total_karma', 0)
        ("")

        self.has_premium: bool = d['is_gold']
        ("")
        self.has_verified_email: bool = d['has_verified_email']
        ("")

        self.is_admin: bool = d['is_employee']
        ("""
            Is a Reddit admin.
            """)
        self.is_a_subreddit_moderator: bool = d['is_mod']
        ("""
            Whether the account is a moderator of any subreddit.
            """)

        self.icon_img: str = d['icon_img']
        ("")

        self.subreddit: User.Subreddit = self.Subreddit(d)
        ("")
        self.me: User.Me = self.Me(d)
        ("""
            Attributes relating to the current user.

            If there is no user context, these values contain nonsense.
            """)


class SuspendedUser(DatamementoPropertiesMixin):
    class Me:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.is_blocked: bool = d['is_blocked']
            ("")

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.name: str = d['name']
        ("")
        self.awardee_karma: int = d['awardee_karma']
        ("")
        self.awarder_karma: int = d['awarder_karma']
        ("")
        self.total_karma: int = d['total_karma']
        ("")

        self.me: SuspendedUser.Me = self.Me(d)
        ("")




from __future__ import annotations
from typing import TYPE_CHECKING, MutableMapping
if TYPE_CHECKING:
    from ..send_params import SendParams
    from ..exchange import Exchange
    from ..requisition import Requisition

from ..handler_ASYNC import Handler
from ..delegating_handler_ASYNC import DelegatingHandler


class DestinationUndeterminedException(Exception):
    pass

class DestinationCollisionException(Exception):
    pass


class Dispatcher(Handler):
    def __init__(self) -> None:
        super().__init__()
        self.destinations: MutableMapping[Requisition, Handler] = {}
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        try:
            dest = self.destinations[p.requisition]
        except KeyError:
            raise DestinationUndeterminedException from None
        return await dest._send(p)

class Demultiplexer(DelegatingHandler):
    def __init__(self, handler: Handler, dispatcher: Dispatcher, destination: Handler) -> None:
        super().__init__(handler)
        self.dispatcher: Dispatcher = dispatcher
        ("")
        self.destination: Handler = destination
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        dests = self.dispatcher.destinations
        reqi = p.requisition
        dest = self.destination
        if dests.get(reqi, dest) is not dest:
            raise DestinationCollisionException
        dests[reqi] = dest
        try:
            xchg = await super()._send(p)
        finally:
            del dests[reqi]
        return xchg

    async def _close(self) -> None:
        await super()._close()
        await self.destination._close()




from typing import Mapping, Union, IO, Tuple

RequestFiles = Mapping[str, Union[str, IO[bytes]]]
ExtendedRequestFiles = Mapping[
    str,
    Union[
        str,
        IO[bytes],
        Tuple[IO[bytes], str],
        Tuple[IO[bytes], str, str],
    ],
]




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Generic, Optional
if TYPE_CHECKING:
    from typing import Iterable, Iterator, Callable, Sequence

from .stubborn_caller_iterator import StubbornCallerIterator
from .unfaltering_chaining_iterator import UnfalteringChainingIterator

T = TypeVar('T')

class CallChunkChainingIterator(Generic[T]):
    """Evaluate call chunks and chain them together."""

    @property
    def current_callable(self) -> Optional[Callable[[], Sequence[T]]]:
        return self.__calling_itr.current

    @current_callable.setter
    def current_callable(self, value: Optional[Callable[[], Sequence[T]]]) -> None:
        self.__calling_itr.current = value

    @property
    def current_iterator(self) -> Iterator[T]:
        return self.__chaining_itr.current_iterator

    @current_iterator.setter
    def current_iterator(self, value: Iterator[T]) -> None:
        self.__chaining_itr.current_iterator = value

    def __init__(self, chunks: Iterable[Callable[[], Sequence[T]]]) -> None:
        self.__chunking_itr: Iterator[Callable[[], Sequence[T]]] = iter(chunks)
        self.__calling_itr: StubbornCallerIterator[Sequence[T]] = StubbornCallerIterator(self.__chunking_itr)
        self.__chaining_itr: UnfalteringChainingIterator[T] = UnfalteringChainingIterator(self.__calling_itr)

    def __iter__(self) -> Iterator[T]:
        return self

    def __next__(self) -> T:
        return next(self.__chaining_itr)

    def get_chunking_iterator(self) -> Iterator[Callable[[], Sequence[T]]]:
        return self.__chunking_itr




from .widget_SYNC import Widget as Widget  # noqa: F401
from .text_area_SYNC import TextAreaWidget as TextAreaWidget  # noqa: F401
from .button_SYNC import ButtonWidget as ButtonWidget  # noqa: F401
from .image_SYNC import ImageWidget as ImageWidget  # noqa: F401
from .community_list_SYNC import CommunityListWidget as CommunityListWidget  # noqa: F401
from .calendar_SYNC import CalendarWidget as CalendarWidget  # noqa: F401
from .post_flair_SYNC import PostFlairWidget as PostFlairWidget  # noqa: F401
from .custom_css_SYNC import CustomCSSWidget as CustomCSSWidget  # noqa: F401
from .community_details_SYNC import CommunityDetailsWidget as CommunityDetailsWidget  # noqa: F401
from .moderator_list_SYNC import ModeratorListWidget as ModeratorListWidget  # noqa: F401
from .rules_SYNC import RulesWidget as RulesWidget  # noqa: F401

from .menu_bar_SYNC import MenuBar as MenuBar  # noqa: F401

from .widget_list_SYNC import WidgetList as WidgetList  # noqa: F401



"""Provides functions that makes functions more decorator-friendly."""

from __future__ import annotations
from typing import TypeVar, Callable

T = TypeVar('T')

def passthru(func: Callable[[T], None]) -> Callable[[T], T]:
    """Make a function that takes an argument and returns `None` return the argument that was given to it."""
    def inner(arg: T) -> T:
        func(arg)
        return arg
    return inner

def passover(func: Callable[[T], object]) -> Callable[[T], T]:
    """Same as `passthru()` but accept any function returning any type instead of just `None`."""
    def inner(arg: T) -> T:
        func(arg)
        return arg
    return inner




from __future__ import annotations
from typing import TYPE_CHECKING, MutableMapping, Protocol, MutableSequence
if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec
    from .connector_ASYNC import Connector

from dataclasses import dataclass

from ...util.imports import load_spec, import_module_from_spec


class NewConnectorFunctionProtocol(Protocol):
    def __call__(self) -> Connector: ...

@dataclass
class TransportInfo:
    adaptor_module_name: str
    name: str
    version: str
    new_connector: NewConnectorFunctionProtocol


def load_transport() -> TransportInfo:
    if not transport_registry:
        for module_spec in transport_module_spec_list:
            try:
                import_module_from_spec(module_spec)
            except ImportError:
                continue
            break
        else:
            raise ModuleNotFoundError('An HTTP transport library needs to be installed.')

    return next(iter(transport_registry.values()))

def new_connector() -> Connector:
    new_connector = load_transport().new_connector
    return new_connector()

def register(
    *,
    adaptor_module_name: str,
    name: str,
    version: str,
    new_connector: NewConnectorFunctionProtocol,
) -> None:
    info = TransportInfo(
        adaptor_module_name=adaptor_module_name,
        name=name,
        version=version,
        new_connector=new_connector,
    )
    transport_registry[adaptor_module_name] = info

transport_module_spec_list: MutableSequence[ModuleSpec] = [
    load_spec('.connectors.httpx_async', __package__),
    load_spec('.connectors.aiohttp', __package__),
]
transport_registry: MutableMapping[str, TransportInfo] = {}




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_ASYNC import LooseComment

from .pull_ASYNC import Pull

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.listing.comment_listing_async_paginator import LooseCommentListingAsyncPaginator

class FrontProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.pull: Pull = Pull(client)
        ("""
            Pull front page submissions.
            """)

    def pull_new_comments(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[LooseCommentListingAsyncPaginator, LooseComment]:
        """Pull new comments from all of Reddit."""
        p = LooseCommentListingAsyncPaginator(self._client, '/comments')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.user_relationship import (
        UserRelationship,
        BannedSubredditUserRelationship,
    )
    from ...models.moderator_list import ModeratorListItem

from ...model_loaders.moderator_list import load_moderator_list_item
from ...model_loaders.user_relationship import load_user_relationship, load_banned_subreddit_user_relation

from .legacy_pull_users_SYNC import LegacyPullUsers

class Legacy:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.pull_users: LegacyPullUsers = LegacyPullUsers(client)
        ("""
            Get redditors that relate to a subreddit.

            .. .RAISES

            :(raises): Same as :meth:`.get_approved_user`.
            """)

    def list_moderators(self, sr: str) -> Sequence[ModeratorListItem]:
        """Get a list of moderators from a subreddit.

        This procedure doesn't work (403 HTTP error) if there is no user context.
        Use :meth:`~.siteprocs.moderation.pull_users_SYNC.PullUsers.moderators` instead.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.moderator_list.ModeratorListItem`]

        .. .RAISES

        :(raises): In addition to :meth:`.get_approved_user`:

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - The current user is banned from the subreddit.
        """
        root = self._client.request('GET', f'/r/{sr}/about/moderators')
        return [load_moderator_list_item(d) for d in root['data']['children']]


    def get_approved_user(self, sr: str, user: str) -> Optional[UserRelationship]:
        """Get information about an approved user in a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.user_relationship.UserRelationship`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `private`:
                The target subreddit is private.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
                The target subreddit does not exist.
            + `403`:
                You don't have access to the subreddit.
            + `404`:
                The specified subreddit name is too long (over 21 characters)
                or contains invalid characters.
        """
        root = self._client.request('GET', f'/r/{sr}/about/contributors', params={'user': user})
        children = root['data']['children']
        return load_user_relationship(children[0]) if children else None

    def get_wiki_contributor(self, sr: str, user: str) -> Optional[UserRelationship]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/r/{sr}/about/wikicontributors', params={'user': user})
        children = root['data']['children']
        return load_user_relationship(children[0]) if children else None

    def get_banned_user(self, sr: str, user: str) -> Optional[BannedSubredditUserRelationship]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/r/{sr}/about/banned', params={'user': user})
        children = root['data']['children']
        return load_banned_subreddit_user_relation(children[0]) if children else None

    def get_muted_user(self, sr: str, user: str) -> Optional[UserRelationship]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/r/{sr}/about/muted', params={'user': user})
        children = root['data']['children']
        return load_user_relationship(children[0]) if children else None

    def get_wiki_banned_user(self, sr: str, user: str) -> Optional[BannedSubredditUserRelationship]:
        """
        Behaves similarly to :meth:`.get_approved_user`.
        """
        root = self._client.request('GET', f'/r/{sr}/about/wikibanned', params={'user': user})
        children = root['data']['children']
        return load_banned_subreddit_user_relation(children[0]) if children else None




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping, Sequence
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from .wiki import (
    WikiPageRevisionAuthorUser as BaseWikiPageRevisionAuthorUser,
    WikiPage as BaseWikiPage,
    WikiPageRevision as BaseWikiPageRevision,
    WikiPageSettings as BaseWikiPageSettings,
)

class WikiPageRevisionAuthorUser(BaseWikiPageRevisionAuthorUser):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")


class WikiPage(BaseWikiPage):
    revision_author: WikiPageRevisionAuthorUser

class WikiPageRevision(BaseWikiPageRevision):
    author: WikiPageRevisionAuthorUser

class WikiPageSettings(BaseWikiPageSettings):
    editors: Sequence[WikiPageRevisionAuthorUser]




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Optional
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .comment import (
    Comment as BaseComment,
    LooseComment as BaseLooseComment,
)

class Comment(BaseComment):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

    def reply(self, text: str) -> Comment:
        return self.client.p.comment.reply(self.id, text)

    def edit_body(self, text: str) -> Comment:
        return self.client.p.comment.edit_body(self.id, text)

    def delete(self) -> None:
        self.client.p.comment.delete(self.id)

    def lock(self) -> None:
        self.client.p.comment.lock(self.id)

    def unlock(self) -> None:
        self.client.p.comment.unlock(self.id)

    def distinguish(self) -> Comment:
        return self.client.p.comment.distinguish(self.id)

    def distinguish_and_sticky(self) -> Comment:
        return self.client.p.comment.distinguish_and_sticky(self.id)

    def undistinguish(self) -> Comment:
        return self.client.p.comment.undistinguish(self.id)

    def approve(self) -> None:
        self.client.p.comment.approve(self.id)

    def remove(self) -> None:
        self.client.p.comment.remove(self.id)

    def remove_spam(self) -> None:
        self.client.p.comment.remove_spam(self.id)

    def apply_removal_reason(self,
            reason_id: Optional[str],
            note: Optional[str] = None) -> None:
        self.client.p.comment.apply_removal_reason(self.id, reason_id, note)

    def send_removal_comment(self,
            title: str,
            message: str) -> Comment:
        return self.client.p.comment.send_removal_comment(self.id, title, message)

    def send_removal_message(self,
            title: str,
            message: str,
            *,
            exposed: bool = False) -> None:
        self.client.p.comment.send_removal_message(self.id, title, message, exposed=exposed)


class LooseComment(Comment, BaseLooseComment):
    pass




from __future__ import annotations

import sys

from .. import __about__
from ..http.transport.reg_SYNC import transport_registry

def get_user_agent(*, module_member: object = None) -> str:
    parts = [
        f"{__about__.__title__}/{__about__.__version__}",
        f"Python/{sys.version.split(None, 1)[0]}",
    ]
    if module_member is not None:
        ti = transport_registry.get(module_member.__module__)
        if ti is not None:
            parts.append(f"{ti.name}/{ti.version}")
    return ' '.join(parts)



"""Mapping wrappers that allow dot access to mapping entires.

Examples:

::

    >>> d = AttributeMappingProxy({'foo': 'bar'})
    >>> d.foo
    'bar'

::

    >>> d = DictRecursiveAttributeMappingProxy({'foo': {'bar': 'baz'}})
    >>> d.foo.bar
    'baz'

::

    >>> d = DictAndListRecursiveAttributeMappingProxy({'foo': [1, 'bar', {'c': 3}]})
    >>> d.foo[2].c
    3
"""

from __future__ import annotations
from typing import Any, TypeVar, Mapping, Iterable, Iterator, IO, Sequence, overload, Union

from pprint import PrettyPrinter

V = TypeVar('V')

class AttributeMappingProxy(Mapping[str, V]):
    __slots__ = ('_store',)
    _store: Mapping[str, V]

    def __init__(self, mapping: Mapping[str, V]) -> None:
        object.__setattr__(self, '_store', mapping)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._store})'

    def __dir__(self) -> Iterable[str]:
        return self.keys()

    def __contains__(self, item: object) -> bool:
        return item in self._store

    def __iter__(self) -> Iterator[str]:
        return iter(self._store)

    def __len__(self) -> int:
        return len(self._store)

    def __getitem__(self, key: str) -> V:
        return self._store[key]

    def __getattr__(self, name: str) -> Any:
        try:
            return self[name]
        except KeyError:
            raise AttributeError(repr(name)) from None

    def __abs__(self) -> Mapping[str, V]:
        return self._store

    def __getstate__(self) -> Mapping[str, V]:
        return self._store

    def __setstate__(self, state: Mapping[str, V]) -> None:
        object.__setattr__(self, '_store', state)

    @staticmethod
    def _pprint(
        printer: PrettyPrinter,
        obj: AttributeMappingProxy[V],
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Mapping[int, Any],
        level: int,
    ) -> None:
        cls_name = obj.__class__.__name__
        stream.write(cls_name + '(')
        printer._format(  # type: ignore[attr-defined]
            dict(obj),
            stream,
            indent + len(cls_name) + 1,
            allowance + 1,
            context,
            level,
        )
        stream.write(')')

    if isinstance(getattr(PrettyPrinter, '_dispatch', None), dict):
        PrettyPrinter._dispatch[__repr__] = _pprint.__func__  # type: ignore[attr-defined]


class MappingRecursiveAttributeMappingProxy(AttributeMappingProxy[V]):
    def __getattr__(self, name: str) -> Any:
        attr = super().__getattr__(name)
        if isinstance(attr, Mapping):
            return MappingRecursiveAttributeMappingProxy(attr)
        return attr

class DictRecursiveAttributeMappingProxy(AttributeMappingProxy[V]):
    def __getattr__(self, name: str) -> Any:
        attr = super().__getattr__(name)
        if isinstance(attr, dict):
            return DictRecursiveAttributeMappingProxy(attr)
        return attr



class ListProxy(Sequence[Any]):
    def __init__(self, data: Sequence[Any]) -> None:
        self._elements: Sequence[Any] = data

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._elements})'

    def __abs__(self) -> Sequence[Any]:
        return self._elements

    def __len__(self) -> int:
        return len(self._elements)

    def __contains__(self, item: object) -> bool:
        return item in self._elements

    def __iter__(self) -> Iterator[Any]:
        return iter(self._elements)

    @overload
    def __getitem__(self, index: int) -> Any: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Any]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Any, Sequence[Any]]:
        item = self._elements[index]
        if isinstance(item, dict):
            return DictAndListRecursiveAttributeMappingProxy(item)
        if isinstance(item, list):
            return self.__class__(item)
        return item

    @staticmethod
    def _pprint(
        printer: PrettyPrinter,
        obj: ListProxy,
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Mapping[int, Any],
        level: int,
    ) -> None:
        cls_name = obj.__class__.__name__
        stream.write(cls_name + '(')
        printer._format(  # type: ignore[attr-defined]
            list(obj),
            stream,
            indent + len(cls_name) + 1,
            allowance + 1,
            context,
            level,
        )
        stream.write(')')

    if isinstance(getattr(PrettyPrinter, '_dispatch', None), dict):
        PrettyPrinter._dispatch[__repr__] = _pprint.__func__  # type: ignore[attr-defined]

class DictAndListRecursiveAttributeMappingProxy(AttributeMappingProxy[V]):
    def __getattr__(self, name: str) -> Any:
        attr = super().__getattr__(name)
        if isinstance(attr, dict):
            return DictAndListRecursiveAttributeMappingProxy(attr)
        elif isinstance(attr, list):
            return ListProxy(attr)
        return attr




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, Any, TypeVar, Mapping
if TYPE_CHECKING:
    from ....client_SYNC import Client

from ...paginator import HasMorePaginator, CursorPaginator
from ....models.moderation_note import (
        ModerationNote,
        ModerationUserNote,
    )
from ....model_loaders.moderation_note import (
    load_moderation_note,
    load_moderation_user_note,
)


T = TypeVar('T', bound='ModerationNote')

class GBaseModerationNotePaginator(HasMorePaginator[T], CursorPaginator[T]):
    def __init__(self,
        client: Client,
        url: str = '/api/mod/notes',
        *,
        limit: Optional[int] = 100,
        cursor: str = '',
        subreddit: str,
        user: str,
        type: Optional[str] = None,
    ) -> None:
        super().__init__(limit=limit)
        self.client: Client = client
        ("")
        self.url: str = url
        ("")
        self._cursor: str = cursor
        self._has_next_page: bool = True
        self._subreddit: str = subreddit
        self._user: str = user
        self._type: Optional[str] = type

    def get_cursor(self) -> str:
        return self._cursor

    def set_cursor(self, value: str) -> None:
        self._cursor = value

    def has_more(self) -> bool:
        return self._has_next_page

    def set_has_more(self, value: bool) -> None:
        self._has_next_page = value

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        if self.limit is not None:
            yield ('count', str(self.limit))

        if self._cursor:
            yield ('before', self._cursor)

        yield ('subreddit', self._subreddit)
        yield ('user', self._user)
        if self._type:
            yield ('filter', self._type)

    def _fetch_data(self) -> Any:
        params = dict(self._generate_params())
        root = self.client.request('GET', self.url, params=params)
        self._cursor = root['end_cursor']
        self._has_next_page = root['has_next_page']
        return root

    def fetch(self) -> Sequence[T]:
        root = self._fetch_data()
        mod_notes = root['mod_notes']
        return [self._load_mod_note(d) for d in mod_notes]

    def _load_mod_note(self, d: Mapping[str, Any]) -> T:
        raise NotImplementedError


class ModerationNotePaginator(GBaseModerationNotePaginator[ModerationNote]):
    def _load_mod_note(self, d: Mapping[str, Any]) -> ModerationNote:
        return load_moderation_note(d)

class ModerationUserNotePaginator(GBaseModerationNotePaginator[ModerationUserNote]):
    def __init__(self,
        client: Client,
        url: str = '/api/mod/notes',
        *,
        limit: Optional[int] = 100,
        cursor: str = '',
        subreddit: str,
        user: str,
        type: Optional[str] = None,
    ) -> None:
        if type != 'NOTE':
            raise ValueError('`type` must be `NOTE`')
        super().__init__(
            client=client,
            url=url,
            cursor=cursor,
            subreddit=subreddit,
            user=user,
            type=type,
        )

    def _load_mod_note(self, d: Mapping[str, Any]) -> ModerationUserNote:
        return load_moderation_user_note(d)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...http.handler_ASYNC import Handler
    from ...http.send_params import SendParams
    from ...http.exchange import Exchange

from ...util.imports import lazy_import
if TYPE_CHECKING:
    import asyncio
else:
    lazy_import % 'asyncio'

from ...http.delegating_handler_ASYNC import DelegatingHandler
from ...util.token_bucket import TokenBucket

class RateLimited(DelegatingHandler):
    def __init__(self, handler: Handler) -> None:
        super().__init__(handler)
        self._tb = TokenBucket(1, 1)

    async def _send(self, p: SendParams) -> Exchange:
        await asyncio.sleep(self._tb.get_cooldown(1))
        self._tb.consume(1)
        return await super()._send(p)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..http.handler_SYNC import Handler
    from ..http.send_params import SendParams
    from ..http.exchange import Exchange

import time

from ..http.delegating_handler_SYNC import DelegatingHandler
from ..util.token_bucket import TokenBucket

class RateLimited(DelegatingHandler):
    def __init__(self, handler: Handler) -> None:
        super().__init__(handler)
        self.remaining: int = 0
        ("")
        self.reset: int = 0
        ("")
        self.used: int = 0
        ("")
        self._delta: float = 0.
        self._timestamp: float = time.monotonic()
        self._burst_control = TokenBucket(5, 1/2)

    def _send(self, p: SendParams) -> Exchange:
        h = self._burst_control.hard_consume(1)
        s = 0.
        if self.remaining <= 1:
            s = self.reset
        else:
            w = self.reset / self.remaining
            if not h or w >= 2:
                # Burst this request, but only if the API didn't
                # want us to wait for more than two seconds.
                # Use 2 because at worst the user is allocated a 2/s (600/300)
                # rate limit when a client credentials grant is used.
                s = w

        time.sleep(s)

        xchg = super()._send(p)

        now = time.monotonic()
        self._delta = now - self._timestamp
        self._timestamp = now

        headers = xchg.response.headers
        if 'x-ratelimit-reset' in headers:
            self.remaining = int(float(headers['x-ratelimit-remaining']))
            self.reset = int(headers['x-ratelimit-reset'])
            self.used = int(headers['x-ratelimit-used'])
        else:
            if self.reset > 0:
                self.remaining -= 1
                self.reset = max(0, self.reset - int(self._delta))
                self.used += 1
            else:
                self.remaining = 300
                self.reset = 600
                self.used = 0

        return xchg




from __future__ import annotations
from typing import TYPE_CHECKING, ClassVar, Any
if TYPE_CHECKING:
    from typing import Mapping
    from ..http.response import Response
    from .token import Token

import re

from ..exceptions import ArgExcMixin

class ArgExc(ArgExcMixin):
    pass


class UnknownTokenType(ArgExc):
    """An exception for when the client receives an unexpected token type.

    This exception class is provided for user code to raise.

    Typically the client will expect that the `token_type` field in a token
    response has the value `bearer`, and the token should be rejected if not.
    See `section 7.1`_ of the OAuth2 specification.

    .. _`section 7.1`: https://datatracker.ietf.org/doc/html/rfc6749#section-7.1
    """

    def __init__(self, arg: object = None, *, token: Token) -> None:
        super().__init__(arg)
        self.token: Token = token

class OAuth2ResponseError(ArgExc):
    """An OAuth2 response error as detailed in the OAuth2 spec.

    For more information see `section 5.2`_ of the OAuth2 specification.

    .. _`section 5.2`: https://tools.ietf.org/html/rfc6749#section-5.2
    """
    ERROR_NAME: ClassVar[str] = ''

    def __init__(self, arg: object = None, *,
            error_name: str = '',
            description: str = '',
            help_uri: str = '') -> None:
        super().__init__(arg)
        self.error_name: str = error_name
        self.description: str = description
        self.help_uri: str = help_uri

    def get_default_message(self) -> str:
        if self.description:
            return repr(self.description)
        return ''

class TokenServerResponseError(OAuth2ResponseError):
    """Error responses that from the token server."""

class TokenServerResponseErrorTypes:
    class InvalidRequest(TokenServerResponseError):
        ERROR_NAME: ClassVar[str] = 'invalid_request'

    class InvalidClient(TokenServerResponseError):
        ERROR_NAME: ClassVar[str] = 'invalid_client'

    class InvalidGrant(TokenServerResponseError):
        ERROR_NAME: ClassVar[str] = 'invalid_grant'

    class UnauthorizedClient(TokenServerResponseError):
        ERROR_NAME: ClassVar[str] = 'unauthorized_client'

    class UnsupportedGrantType(TokenServerResponseError):
        ERROR_NAME: ClassVar[str] = 'unsupported_grant_type'

    class InvalidScope(TokenServerResponseError):
        ERROR_NAME: ClassVar[str] = 'invalid_scope'

class UnrecognizedTokenServerResponseError(TokenServerResponseError):
    pass

token_server_response_error_by_error_name: Mapping[str, type[TokenServerResponseError]] = {
    cls.ERROR_NAME: cls
    for cls in [
        TokenServerResponseError,
        TokenServerResponseErrorTypes.InvalidRequest,
        TokenServerResponseErrorTypes.InvalidClient,
        TokenServerResponseErrorTypes.InvalidGrant,
        TokenServerResponseErrorTypes.UnauthorizedClient,
        TokenServerResponseErrorTypes.UnsupportedGrantType,
        TokenServerResponseErrorTypes.InvalidScope,
    ]
}

def raise_for_token_server_response_error(json_dict: Any) -> None:
    error_name = json_dict.get('error')
    if error_name is None:
        return

    cls = token_server_response_error_by_error_name.get(error_name, UnrecognizedTokenServerResponseError)
    raise cls(
        error_name=error_name,
        description=json_dict.get('error_description', ''),
        help_uri=json_dict.get('error_uri', ''),
    )

class ResourceServerResponseError(OAuth2ResponseError):
    """Error responses that from the resource server (i.e., the API)."""

class ResourceServerResponseErrorTypes:
    class InvalidRequest(ResourceServerResponseError):
        ERROR_NAME: ClassVar[str] = 'invalid_request'

    class InvalidToken(ResourceServerResponseError):
        ERROR_NAME: ClassVar[str] = 'invalid_token'

    class InsufficientScope(ResourceServerResponseError):
        ERROR_NAME: ClassVar[str] = 'insufficient_scope'

class UnrecognizedResourceServerResponseError(ResourceServerResponseError):
    pass

resource_server_response_error_by_error_name: Mapping[str, type[ResourceServerResponseError]] = {
    cls.ERROR_NAME: cls
    for cls in [
        ResourceServerResponseError,
        ResourceServerResponseErrorTypes.InvalidRequest,
        ResourceServerResponseErrorTypes.InvalidToken,
        ResourceServerResponseErrorTypes.InsufficientScope,
    ]
}

def raise_for_resource_server_response_error(json_dict: Any) -> None:
    error_name = json_dict.get('error')
    if error_name is None:
        return

    cls = resource_server_response_error_by_error_name.get(
            error_name, UnrecognizedResourceServerResponseError)
    raise cls(
        error_name=error_name,
        description=json_dict.get('error_description', ''),
        help_uri=json_dict.get('error_uri', ''),
    )

_auth_param_pattern = r'''(?P<key>(\w+))=((?P<q>\")(?P<value>([^\"]*))(?P=q)|(?P=value))'''
_auth_param_regex = re.compile(_auth_param_pattern)

def extract_www_authenticate_auth_params(resp: Response) -> Mapping[str, str]:
    try:
        www_authenticate = resp.headers['WWW-Authenticate']
    except KeyError:
        return {}

    # Parsing the WWW-Authenticate header in a RFC-2617 spec compliant way is a daunting task.
    # Just use regex for now.
    return {m['key']: m['value'] for m in _auth_param_regex.finditer(www_authenticate)}




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_ASYNC import Submission
    from ...models.comment_ASYNC import LooseComment
    from ...models.subreddit_ASYNC import Subreddit
    from ..stream_ASYNC import IStandardStreamEventSubject
    from ...pagination.async_paginator import CursorAsyncPaginator

from ..stream_ASYNC import Stream


def submission_stream_extractor(x: Submission) -> object:
    return x.id

def get_submission_stream_paginator(client: Client, sr: str) -> CursorAsyncPaginator[Submission]:
    return client.p.subreddit.pull.new(sr).get_paginator()

def make_submission_stream(paginator: CursorAsyncPaginator[Submission], past: Optional[Iterable[Submission]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[Submission]:
    return Stream(paginator, submission_stream_extractor, past=past, seen=seen)

def create_submission_stream(client: Client, sr: str) -> IStandardStreamEventSubject[Submission]:
    return make_submission_stream(get_submission_stream_paginator(client, sr))


def comment_stream_extractor(x: LooseComment) -> object:
    return x.id

def get_comment_stream_paginator(client: Client, sr: str) -> CursorAsyncPaginator[LooseComment]:
    return client.p.subreddit.pull_new_comments(sr).get_paginator()

def make_comment_stream(paginator: CursorAsyncPaginator[LooseComment], past: Optional[Iterable[LooseComment]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[LooseComment]:
    return Stream(paginator, comment_stream_extractor, past=past, seen=seen)

def create_comment_stream(client: Client, sr: str) -> IStandardStreamEventSubject[LooseComment]:
    return make_comment_stream(get_comment_stream_paginator(client, sr))


def subreddit_stream_extractor(x: Subreddit) -> object:
    return x.id

def get_subreddit_stream_paginator(client: Client) -> CursorAsyncPaginator[Subreddit]:
    return client.p.subreddit.pulls.new().get_paginator()

def make_subreddit_stream(paginator: CursorAsyncPaginator[Subreddit], past: Optional[Iterable[Subreddit]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[Subreddit]:
    return Stream(paginator, subreddit_stream_extractor, past=past, seen=seen)

def create_subreddit_stream(client: Client) -> IStandardStreamEventSubject[Subreddit]:
    return make_subreddit_stream(get_subreddit_stream_paginator(client))




from __future__ import annotations
from typing import TYPE_CHECKING, IO, Optional, Iterable, Protocol
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_style_asset_upload_lease import SubredditStyleAssetUploadLease

import os.path as op
from functools import cached_property

from ...model_loaders.subreddit_style_asset_upload_lease import load_subreddit_style_asset_upload_lease
from ...http.util.guess_filename_mimetype import guess_filename_mimetype

class SubredditStyleNewProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    class BannerUploading:
        def __init__(self, outer: SubredditStyleNewProcedures) -> None:
            self._client = outer._client

        async def _obtain_upload_lease(self,
            *,
            sr: str,
            filepath: str,
            mimetype: Optional[str] = None,
            imagetype: str,
        ) -> SubredditStyleAssetUploadLease:
            if mimetype is None:
                mimetype = guess_filename_mimetype(filepath)
            result = await self._client.request('POST', f'/api/v1/style_asset_upload_s3/{sr}',
                    data={'filepath': filepath, 'mimetype': mimetype, 'imagetype': imagetype})
            return load_subreddit_style_asset_upload_lease(result)

        async def obtain_banner_upload_lease(self,
            *,
            sr: str,
            filepath: str,
            mimetype: Optional[str] = None,
        ) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filepath=filepath, mimetype=mimetype, imagetype='bannerBackgroundImage')

        async def obtain_banner_overlay_upload_lease(self,
            *,
            sr: str,
            filepath: str,
            mimetype: Optional[str] = None,
        ) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filepath=filepath, mimetype=mimetype, imagetype='bannerPositionedImage')

        async def obtain_banner_overlay_hover_upload_lease(self,
            *,
            sr: str,
            filepath: str,
            mimetype: Optional[str] = None,
        ) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filepath=filepath, mimetype=mimetype, imagetype='secondaryBannerPositionedImage')

        async def obtain_mobile_banner_upload_lease(self,
            *,
            sr: str,
            filepath: str,
            mimetype: Optional[str] = None,
        ) -> SubredditStyleAssetUploadLease:
            return await self._obtain_upload_lease(sr=sr, filepath=filepath, mimetype=mimetype, imagetype='mobileBannerImage')

        async def deposit_file(self,
            file: IO[bytes],
            upload_lease: SubredditStyleAssetUploadLease,
            *,
            timeout: float = 1000,
        ) -> None:
            resp = await self._client.http.request('POST', upload_lease.endpoint,
                    data=upload_lease.fields, files={'file': file}, timeout=timeout)
            resp.ensure_successful_status()

        class ObtainUploadLeaseFunction(Protocol):
            async def __call__(self,
                *,
                sr: str,
                filepath: str,
                mimetype: Optional[str] = None,
            ) -> SubredditStyleAssetUploadLease: ...

        async def _upload(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
            obtain_upload_lease: ObtainUploadLeaseFunction,
        ) -> SubredditStyleAssetUploadLease:
            if filepath is None:
                filepath = op.basename(getattr(file, 'name', ''))
                if not filepath:
                    raise ValueError("the `filepath` parameter must be explicitly specified if the file object has no `name` attribute.")
            upload_lease = await obtain_upload_lease(sr=sr, filepath=filepath)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

        async def upload_banner(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> SubredditStyleAssetUploadLease:
            return await self._upload(file=file, sr=sr, filepath=filepath, timeout=timeout, obtain_upload_lease=self.obtain_banner_upload_lease)

        async def upload_banner_overlay(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> SubredditStyleAssetUploadLease:
            return await self._upload(file=file, sr=sr, filepath=filepath, timeout=timeout, obtain_upload_lease=self.obtain_banner_overlay_upload_lease)

        async def upload_banner_overlay_hover(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> SubredditStyleAssetUploadLease:
            return await self._upload(file=file, sr=sr, filepath=filepath, timeout=timeout, obtain_upload_lease=self.obtain_banner_overlay_hover_upload_lease)

        async def upload_mobile_banner(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> SubredditStyleAssetUploadLease:
            return await self._upload(file=file, sr=sr, filepath=filepath, timeout=timeout, obtain_upload_lease=self.obtain_mobile_banner_upload_lease)

    banner_uploading: cached_property[BannerUploading] = cached_property(BannerUploading)

    async def modify_banner_settings(self,
        sr: str,
        *,
        banner_size: Optional[str] = None,
        banner_background_color: Optional[str] = None,
        banner_image_url: Optional[str] = None,
        banner_image_display: Optional[str] = None,
        banner_overlay_image_url: Optional[str] = None,
        banner_overlay_image_position: Optional[str] = None,
        banner_overlay_hover_image_url: Optional[str] = None,
        mobile_banner_image_url: Optional[str] = None,
    ) -> None:
        """Set banner images in a subreddit.

        Parameters set to `None` are ignored.

        Use an empty string to set a setting to its default.

        .. .PARAMETERS

        :param `Optional[str]` banner_size:
            Either: `small`, `medium`, `large`.

            Empty string or any other value defaults to `small`.
        :param `Optional[str]` banner_background_color:
            A hex color.

            Empty string or any other value defaults to `#33a8ff`.
        :param `Optional[str]` banner_image_url:
            Location of the a banner image.

            Use empty string to remove the image.
        :param `Optional[str]` banner_image_display:
            Either: `cover`, `tiled`.

            Empty string or any other value defaults to `cover`.
        :param `Optional[str]` banner_overlay_image_url:
            Location of the a banner overlay image.

            Use empty string to remove the image.
        :param `Optional[str]` banner_overlay_image_position:
            Either: `left`, `centered`, `right`.

            Empty string or any other value defaults to `left`.
        :param `Optional[str]` banner_overlay_hover_image_url:
            The location of the a banner overlay hover image.

            Use empty string to remove the image.
        :param `Optional[str]` mobile_banner_image_url:
            The location of the a mobile banner image.

            Use empty string to remove the image.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        def g() -> Iterable[tuple[str, str]]:
            if banner_size is not None:
                yield ('bannerHeight', banner_size)
            if banner_background_color is not None:
                yield ('bannerBackgroundColor', banner_background_color)
            if banner_image_url is not None:
                yield ('bannerBackgroundImage', banner_image_url)
            if banner_image_display is not None:
                v = {
                    '': '',
                    'fill': 'cover',
                    'tile': 'tiled',
                }[banner_image_display]
                yield ('bannerBackgroundImagePosition', v)
            if banner_overlay_image_url is not None:
                yield ('bannerPositionedImage', banner_overlay_image_url)
            if banner_overlay_image_position is not None:
                v = {
                    '': '',
                    'left': 'left',
                    'center': 'centered',
                    'right': 'right',
                }[banner_overlay_image_position]
                yield ('bannerPositionedImagePosition', v)
            if banner_overlay_hover_image_url is not None:
                yield ('secondaryBannerPositionedImage', banner_overlay_hover_image_url)
            if mobile_banner_image_url is not None:
                yield ('mobileBannerImage', mobile_banner_image_url)

        await self._client.request('PATCH', f'/api/v1/structured_styles/{sr}', data=dict(g()))




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_ASYNC import LooseComment
    from ...models.submission_ASYNC import Submission

from functools import cached_property

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.user.async1 import (
    OverviewListingAsyncPaginator,
    CommentsListingAsyncPaginator,
    SubmittedListingAsyncPaginator,
    GildedListingAsyncPaginator,
    UpvotedListingAsyncPaginator,
    DownvotedListingAsyncPaginator,
    HiddenListingAsyncPaginator,
    SavedListingAsyncPaginator,
    SavedSubmissionsListingAsyncPaginator,
    SavedCommentsListingAsyncPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    def overview(self, name: str, amount: Optional[int] = None, *,
            sort: str = 'new',
            ) -> ImpartedPaginatorChainingAsyncIterator[OverviewListingAsyncPaginator, object]:
        p = OverviewListingAsyncPaginator(self._client, f'/user/{name}/overview')
        p.sort = sort
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def comments(self, name: str, amount: Optional[int] = None, *,
            sort: str = 'new',
            ) -> ImpartedPaginatorChainingAsyncIterator[CommentsListingAsyncPaginator, LooseComment]:
        p = CommentsListingAsyncPaginator(self._client, f'/user/{name}/comments')
        p.sort = sort
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def submitted(self, name: str, amount: Optional[int] = None, *,
            sort: str = 'hot',
            ) -> ImpartedPaginatorChainingAsyncIterator[SubmittedListingAsyncPaginator, Submission]:
        p = SubmittedListingAsyncPaginator(self._client, f'/user/{name}/submitted')
        p.sort = sort
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def awards_received(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[GildedListingAsyncPaginator, object]:
        p = GildedListingAsyncPaginator(self._client, f'/user/{name}/gilded')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def awards_given(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[GildedListingAsyncPaginator, object]:
        p = GildedListingAsyncPaginator(self._client, f'/user/{name}/gilded/given')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def upvoted(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[UpvotedListingAsyncPaginator, Submission]:
        p = UpvotedListingAsyncPaginator(self._client, f'/user/{name}/upvoted')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def downvoted(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[DownvotedListingAsyncPaginator, Submission]:
        p = DownvotedListingAsyncPaginator(self._client, f'/user/{name}/downvoted')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def hidden(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[HiddenListingAsyncPaginator, Submission]:
        p = HiddenListingAsyncPaginator(self._client, f'/user/{name}/hidden')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    class Saved:
        def __init__(self, outer: Pull) -> None:
            self._client = outer._client

        def __call__(self, name: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SavedListingAsyncPaginator, object]:
            p = SavedListingAsyncPaginator(self._client, f'/user/{name}/saved')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def submissions(self, name: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SavedSubmissionsListingAsyncPaginator, Submission]:
            p = SavedSubmissionsListingAsyncPaginator(self._client, f'/user/{name}/saved')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

        def comments(self, name: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingAsyncIterator[SavedCommentsListingAsyncPaginator, LooseComment]:
            p = SavedCommentsListingAsyncPaginator(self._client, f'/user/{name}/saved')
            return ImpartedPaginatorChainingAsyncIterator(p, amount)

    saved: cached_property[Saved] = cached_property(Saved)



"""Reddit API specific constants."""

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Set

AUTHORIZATION_BASE_URL: str = "https://www.reddit.com"
AUTHORIZATION_ENDPOINT_PATH: str = "/api/v1/authorize"
AUTHORIZATION_URL: str = AUTHORIZATION_BASE_URL + AUTHORIZATION_ENDPOINT_PATH
MOBILE_AUTHORIZATION_ENDPOINT_PATH: str = AUTHORIZATION_ENDPOINT_PATH + ".compact"
MOBILE_AUTHORIZATION_URL: str = AUTHORIZATION_BASE_URL + MOBILE_AUTHORIZATION_ENDPOINT_PATH
TOKEN_OBTAINMENT_ENDPOINT_PATH: str = "/api/v1/access_token"
TOKEN_OBTAINMENT_URL: str = AUTHORIZATION_BASE_URL + TOKEN_OBTAINMENT_ENDPOINT_PATH
TOKEN_REVOCATION_ENDPOINT_PATH: str = "/api/v1/revoke_token"
TOKEN_REVOCATION_URL: str = AUTHORIZATION_BASE_URL + TOKEN_REVOCATION_ENDPOINT_PATH

RESOURCE_BASE_URL: str = "https://oauth.reddit.com"

TRUSTED_ORIGINS: Set[str] = {
    RESOURCE_BASE_URL,
    "https://gateway.reddit.com",
}




from __future__ import annotations
from typing import Any, Mapping

from datetime import datetime, timezone

from ..models.moderation_note import ModerationNote, ModerationActionNote, ModerationUserNote

def load_moderation_note(d: Mapping[str, Any]) -> ModerationNote:
    if d['type'] == 'NOTE':
        return load_moderation_user_note(d)
    return load_moderation_action_note(d)

def load_moderation_action_note(d: Mapping[str, Any]) -> ModerationActionNote:
    return ModerationActionNote(
        d=d,
        uuid=d['id'].partition('_')[-1],
        unixtime=(unixtime := int(d['created_at'])),
        datetime=datetime.fromtimestamp(unixtime, timezone.utc),
        label=d['type'],
        subreddit_id=int(d['subreddit_id'].partition('_')[-1], 36),
        subreddit=d['subreddit'],
        agent_id=int(d['operator_id'].partition('_')[-1], 36),
        agent=d['operator'],
        target_id=int(d['user_id'].partition('_')[-1], 36),
        target=d['user'],
        action=d['mod_action_data']['action'],
    )

def load_moderation_user_note(d: Mapping[str, Any]) -> ModerationUserNote:
    d_user_note_data = d['user_note_data']

    anchor_submission_id = None
    anchor_comment_id = None
    reddit_id: str = d_user_note_data['reddit_id'] or ''
    if reddit_id:
        thing, _, id36 = reddit_id.partition('_')
        idn = int(id36, 36)
        if thing == 't3':
            anchor_submission_id = idn
        elif thing == 't1':
            anchor_comment_id = idn

    return ModerationUserNote(
        d=d,
        uuid=d['id'].partition('_')[-1],
        unixtime=(unixtime := int(d['created_at'])),
        datetime=datetime.fromtimestamp(unixtime, timezone.utc),
        subreddit_id=int(d['subreddit_id'].partition('_')[-1], 36),
        subreddit=d['subreddit'],
        agent_id=int(d['operator_id'].partition('_')[-1], 36),
        agent=d['operator'],
        target_id=int(d['user_id'].partition('_')[-1], 36),
        target=d['user'],
        note=d_user_note_data['note'],
        label=d_user_note_data['label'] or '',
        has_anchor=bool(reddit_id),
        anchor_submission_id=anchor_submission_id,
        anchor_comment_id=anchor_comment_id,
    )




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.submission_SYNC import Submission

from functools import cached_property

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.wiki_sync1 import WikiPageRevisionsPaginator
from ...pagination.paginators.listing.submission_listing_paginator import SubmissionListingPaginator
from ...models.wiki_SYNC import WikiPage, WikiPageRevision, WikiPageSettings
from ...model_loaders.wiki_SYNC import load_wiki_page, load_wiki_page_settings

class WikiProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    class Page:
        def __init__(self, outer: WikiProcedures) -> None:
            self._outer = outer
            self._client = outer._client

        def read(self, sr: str, page: str) -> WikiPage:
            """Get a wiki page.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:

            .. .RETURNS

            :rtype: :class:`~.models.wiki_SYNC.WikiPage`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `PAGE_NOT_FOUND`:
                    The specified wiki page does not exist.
                + `INVALID_REVISION`:
                    The specified revision UUID does not exist.
                + `private`:
                    You do not have access to the specified subreddit; it is private.
                + `banned`:
                    You do not have access to the specified subreddit; it is banned.
            """
            root = self._client.request('GET', f'/r/{sr}/wiki/{page}')
            return load_wiki_page(root['data'], self._client)

        def edit(self, sr: str, page: str, body: str, *, message: str = '') -> None:
            """Edit a wiki page.

            If the specifed content matches the current content of the wiki page,
            the edit will not go through.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:
            :param `str` body:
            :param `str` message:
                A commit message.

            .. .RETURNS

            :rtype: `None`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `WIKI_CREATE_ERROR`:
                    The specified wiki page does not exist.
                + `private`:
                    You do not have access to the specified subreddit; it is private.
                + `banned`:
                    You do not have access to the specified subreddit; it is banned.
            """
            self._client.request('POST', f'/r/{sr}/api/wiki/edit',
                    data={'page': page, 'content': body, 'reason': message})

        def revert(self, sr: str, page: str, revision: str) -> None:
            """Revert a wiki page to a previous revision.

            This creates a new edit with content matching that of the specified revision.

            The revision message will look something like 'reverted back 53 minutes'.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:
            :param `str` revision:

            .. .RETURNS

            :rtype: `None`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `INVALID_REVISION`:
                    The specified revision UUID does not exist.
            """
            self._client.request('POST', f'/r/{sr}/api/wiki/revert',
                    data={'page': page, 'revision': revision})

        def revisions(self, sr: str, page: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[WikiPageRevisionsPaginator, WikiPageRevision]:
            """Get wiki page revision log entries.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:
            :param `Optional[int]` amount:

            .. .RETURNS

            :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`[:class:`~.pagination.paginators.wiki_sync1.WikiPageRevisionsPaginator`, :class:`~.models.wiki_SYNC.WikiPageRevision`]

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `WIKI_DISABLED`:
                    The specified subreddit does not have wikis enabled.
                + `PAGE_NOT_FOUND`:
                    The specified wiki page does not exist.
                + `private`:
                    You do not have access to the specified subreddit; it is private.
                + `banned`:
                    You do not have access to the specified subreddit; it is banned.
            """
            p = WikiPageRevisionsPaginator(self._client, f'/r/{sr}/wiki/revisions/{page}')
            return ImpartedPaginatorChainingIterator(p, amount)

        def discussions(self, sr: str, page: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SubmissionListingPaginator, Submission]:
            """Get link submissions linking to a particular wiki page.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:
            :param `Optional[int]` amount:

            .. .RETURNS

            :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`[:class:`~.pagination.paginators.wiki_sync1.WikiPageRevisionsPaginator`, :class:`~.models.wiki_SYNC.WikiPageRevision`]

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `WIKI_DISABLED`:
                    The specified subreddit does not have wikis enabled.
                + `PAGE_NOT_FOUND`:
                    The specified wiki page does not exist.
                + `private`:
                    You do not have access to the specified subreddit; it is private.
                + `banned`:
                    You do not have access to the specified subreddit; it is banned.
            """
            p = SubmissionListingPaginator(self._client, f'/r/{sr}/wiki/discussions/{page}')
            return ImpartedPaginatorChainingIterator(p, amount)

        def get_settings(self, sr: str, page: str) -> WikiPageSettings:
            """Retrieve the current permission settings for a wiki page.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:

            .. .RETURNS

            :rtype: `~.models.wiki_SYNC.WikiPageSettings`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `MOD_REQUIRED`:
                    You are not a moderator of the specified subreddit.
                + `WIKI_DISABLED`:
                    The specified subreddit does not have wikis enabled.
                + `PAGE_NOT_FOUND`:
                    The specified wiki page does not exist.
                + `private`:
                    You do not have access to the specified subreddit; it is private.
                + `banned`:
                    You do not have access to the specified subreddit; it is banned.
            """
            root = self._client.request('GET', f'/r/{sr}/wiki/settings/{page}')
            return load_wiki_page_settings(root['data'], self._client)

        def set_settings(self, sr: str, page: str, *, permlevel: int, indexed: bool) -> WikiPageSettings:
            """Update the permissions and visibility of a particular wiki page.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:
            :param `int` permlevel:
                Permission level indicating who can edit this wiki page.

                "who can edit this page?"

                `0`: "use subreddit wiki permissions"
                `1`: "only approved wiki contributors for this page may edit"
                `2`: "only mods may edit and view"
            :param `bool` indexed:
                Whether this wiki page is to be indexed on the wiki page list.

            .. .RETURNS

            :rtype: `~.models.wiki_SYNC.WikiPageSettings`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `WIKI_DISABLED`:
                    The specified subreddit does not have wikis enabled.
                + `PAGE_NOT_FOUND`:
                    The specified wiki page does not exist.
                + `private`:
                    You do not have access to the specified subreddit; it is private.
                + `banned`:
                    You do not have access to the specified subreddit; it is banned.
            :raises redditwarp.http.exceptions.StatusCodeException:
                + `302`:
                    You do not have permission.
            """
            root = self._client.request('POST', f'/r/{sr}/wiki/settings/{page}',
                    data={'permlevel': str(permlevel), 'listed': '01'[indexed]})
            return load_wiki_page_settings(root['data'], self._client)

        def add_editor(self, sr: str, page: str, username: str) -> None:
            """Add a user as an editor for a wiki page.

            If the user is already added, it is treated as a success.

            .. .PARAMETERS

            :param `str` sr:
            :param `str` page:
            :param `str` username:

            .. .RETURNS

            :rtype: `None`

            .. .RAISES

            :raises redditwarp.exceptions.RedditError:
                + `WIKI_DISABLED`:
                    The specified subreddit does not have wikis enabled.
                + `UNKNOWN_USER`:
                    The specified user does not exist.
                + `private`:
                    You do not have access to the specified subreddit; it is private.
                + `banned`:
                    You do not have access to the specified subreddit; it is banned.
            """
            self._client.request('POST', f'/r/{sr}/api/wiki/alloweditor/add', data={'username': username})

        def remove_editor(self, sr: str, page: str, username: str) -> None:
            """Remove a user as an editor for a wiki page.

            Behaves similarly to :meth:`.add_editor`.
            """
            self._client.request('POST', f'/r/{sr}/api/wiki/alloweditor/del', data={'username': username})

    page: cached_property[Page] = cached_property(Page)

    def revisions(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[WikiPageRevisionsPaginator, WikiPageRevision]:
        """Get a revision log for all wiki pages.

        .. .PARAMETERS

        :param `str` sr:
        :param `Optional[int]` amount:

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`[:class:`~.pagination.paginators.wiki_sync1.WikiPageRevisionsPaginator`, :class:`~.models.wiki_SYNC.WikiPageRevision`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `WIKI_DISABLED`:
                The specified subreddit does not have wikis enabled.
            + `private`:
                You do not have access to the specified subreddit; it is private.
            + `banned`:
                You do not have access to the specified subreddit; it is banned.
        """
        p = WikiPageRevisionsPaginator(self._client, f'/r/{sr}/wiki/revisions')
        return ImpartedPaginatorChainingIterator(p, amount)

    def list_pages(self, sr: str) -> Sequence[str]:
        """Get a list of wiki pages in a subreddit.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `Sequence`\\[`str`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `private`:
                You do not have access to the specified subreddit; it is private.
            + `banned`:
                You do not have access to the specified subreddit; it is banned.
        """
        root = self._client.request('GET', f'/r/{sr}/wiki/pages')
        return root['data']

    def toggle_revision_hidden(self, sr: str, page: str, revision: str) -> bool:
        """Toggle the public visibility of a wiki page revision.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` page:
        :param `str` revision:

        .. .RETURNS

        :rtype: `Sequence`\\[`str`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `PAGE_NOT_FOUND`:
                The specified wiki page does not exist.
            + `INVALID_REVISION`:
                The specified revision UUID does not exist.
            + `private`:
                You do not have access to the specified subreddit; it is private.
            + `banned`:
                You do not have access to the specified subreddit; it is banned.
        """
        root = self._client.request('POST', f'/r/{sr}/api/wiki/hide',
                data={'page': page, 'revision': revision})
        return root['status']




from __future__ import annotations
from typing import TypeVar, Sequence, Generic, AsyncIterator, Optional

T = TypeVar('T')

# Abstract classes

class AsyncPaginator(Generic[T]):
    def __init__(self, *, limit: Optional[int] = None) -> None:
        self.limit: Optional[int] = limit
        ("")

    async def __aiter__(self) -> AsyncIterator[Sequence[T]]:
        while page := await self.fetch():
            yield page

    async def fetch(self) -> Sequence[T]:
        raise NotImplementedError

class OffsetAsyncPaginator(AsyncPaginator[T]):
    def __init__(self, *, limit: Optional[int] = None, offset: Optional[int] = None) -> None:
        super().__init__(limit=limit)
        self.offset: Optional[int] = offset
        ("")

class CursorAsyncPaginator(AsyncPaginator[T]):
    def get_cursor(self) -> str:
        raise NotImplementedError

    def set_cursor(self, value: str) -> None:
        raise NotImplementedError


# Interfaces

class Bidirectional:
    direction: bool

class Resettable:
    def reset(self) -> None:
        raise NotImplementedError

class HasMore:
    def has_more(self) -> bool:
        raise NotImplementedError

    def set_has_more(self, value: bool) -> None:
        raise NotImplementedError


# Mixins

class HasMoreAsyncPaginator(HasMore, AsyncPaginator[T]):
    async def __aiter__(self) -> AsyncIterator[Sequence[T]]:
        if page := await self.fetch():
            yield page
            while self.has_more() and (page := await self.fetch()):
                yield page




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .types import ClientCredentials
    from ..http.http_client_ASYNC import HTTPClient

from .utils import apply_basic_auth

class TokenRevocationClient:
    """Makes token revocation requests."""

    def __init__(self, http: HTTPClient, url: str,
            client_creds: ClientCredentials) -> None:
        """
        .. .PARAMETERS

        :param http: An instance of an HTTP client to make the requests.
        :type http: :class:`~.http.http_client_ASYNC.HTTPClient`
        :param url: The URL of the revocation endpoint.
        :type url: str
        :param client_creds: The client credentials to be used in the revocation request.
        :type client_creds: :data:`~.types.ClientCredentials`
        """
        self.http: HTTPClient = http
        ("")
        self.url: str = url
        ("")
        self.client_creds: ClientCredentials = client_creds
        ("")

    async def revoke_token(self, token: str, token_type_hint: str = '') -> None:
        """Makes a revocation request for a token.

        .. .PARAMETERS

        :param token: The token to be revoked.
        :type token: str
        :param token_type_hint: A hint about the type of the token
            (e.g. `access_token`, `refresh_token`).
            This is supposed to make the operation slightly quicker for
            the server. If the given hint is invalid then the server
            should still attempt to resolve and revoke the token.
        :type token_type_hint: str
        """
        data = {'token': token}
        if token_type_hint:
            data['token_type_hint'] = token_type_hint

        headers: dict[str, str] = {}
        apply_basic_auth(headers, *self.client_creds)
        resp = await self.http.request('POST', self.url, headers=headers, data=data)
        resp.ensure_successful_status()

    async def revoke_access_token(self, token: str) -> None:
        """Makes a revocation request for an access token.

        Calls `self.revoke_token(token, 'access_token')`.

        .. .PARAMETERS

        :param token: The access token to be revoked.
        :type token: str
        """
        await self.revoke_token(token, 'access_token')

    async def revoke_refresh_token(self, token: str) -> None:
        """Makes a revocation request for a refresh token.

        Calls `self.revoke_token(token, 'refresh_token')`.

        .. .PARAMETERS

        :param token: The refresh token to be revoked.
        :type token: str
        """
        await self.revoke_token(token, 'refresh_token')




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .send_params import SendParams
    from .exchange import Exchange


class Handler:
    async def _send(self, p: SendParams) -> Exchange:
        raise NotImplementedError

    async def _close(self) -> None:
        pass




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.modmail_ASYNC import ConversationInfo, Message

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.modmail_async1 import ModmailConversationMessageAsyncPaginator

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    def _mailbox(self,
        mailbox: str,
        amount: Optional[int] = None,
        *,
        subreddit_names: Sequence[str] = (),
        sort: str = '',
    ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        p = ModmailConversationMessageAsyncPaginator(self._client, '/api/mod/conversations', mailbox, subreddit_names, sort)
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def all(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('all', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def inbox(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('inbox', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def new(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('new', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def in_progress(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('inprogress', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def archived(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('archived', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def filtered(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('filtered', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def appeals(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('appeals', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def join_requests(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('join_requests', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def highlighted(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('highlighted', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def mod(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('mod', amount=amount, subreddit_names=subreddit_names, sort=sort)

    def notifications(self, amount: Optional[int] = None, *, subreddit_names: Sequence[str] = (), sort: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModmailConversationMessageAsyncPaginator, tuple[ConversationInfo, Message]]:
        return self._mailbox('notifications', amount=amount, subreddit_names=subreddit_names, sort=sort)




from __future__ import annotations
from typing import Mapping, Any

class UserSummary:
    def __init__(self, d: Mapping[str, Any], id36: str) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id36: str = id36
        ("""
            ID of the user account as a base 36 number.
            """)
        self.id: int = int(id36, 36)
        ("""
            ID of the user account.
            """)
        self.name: str = d['name']
        ("""
            Name of the user.
            """)
        self.created_ut: int = int(d['created_utc'])
        ("""
            Unix timestamp of when the user account was created.
            """)
        self.post_karma: int = d['link_karma']
        ("")
        self.comment_karma: int = d['comment_karma']
        ("")
        self.profile_img: str = d['profile_img']
        ("")
        self.profile_color: str = d['profile_color']
        ("")
        self.subreddit_nsfw: bool = d['profile_over_18']
        ("""
            Whether the user's user subreddit is NSFW.
            """)



"""General tree node data structure."""

from __future__ import annotations
from typing import TypeVar, Generic, Sequence

from dataclasses import dataclass


TValue_co = TypeVar('TValue_co', covariant=True)
TChild_co = TypeVar('TChild_co', covariant=True)

@dataclass(repr=False, eq=False, frozen=True)
class TreeNode(Generic[TValue_co, TChild_co]):
    value: TValue_co
    children: Sequence[TChild_co]

RecursiveTreeNode = TreeNode[TValue_co, 'RecursiveTreeNode[TValue_co]']




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping
if TYPE_CHECKING:
    from ..http.handler_SYNC import Handler
    from ..http.send_params import SendParams
    from ..http.exchange import Exchange

from urllib.parse import urlsplit

from ..http.delegating_handler_SYNC import DelegatingHandler

class DirectByOrigin(DelegatingHandler):
    def __init__(self,
        handler: Handler,
        directions: Mapping[str, Handler],
    ) -> None:
        super().__init__(handler)
        self.directions: Mapping[str, Handler] = directions
        ("")

    def _send(self, p: SendParams) -> Exchange:
        o = urlsplit(p.requisition.url)
        origin = f"{o.scheme}://{o.netloc}"
        direction = self.directions.get(origin)
        if direction is not None:
            return direction._send(p)
        return super()._send(p)




from __future__ import annotations
from typing import AsyncIterator

import asyncio
from asyncio.queues import Queue


async def flow(*streams: AsyncIterator[float]) -> None:
    await flow_parallel(*streams)

async def flow_series(*streams: AsyncIterator[float]) -> None:
    loop = asyncio.get_running_loop()

    aq: Queue[AsyncIterator[float]] = Queue()
    count = 0
    for aitr in streams:
        aq.put_nowait(aitr)
        count += 1

    while count > 0:
        aitr = await aq.get()
        try:
            try:
                t = await aitr.__anext__()
            except StopAsyncIteration:
                count -= 1
            else:
                loop.call_later(t, (lambda: aq.put_nowait(aitr)))
        finally:
            aq.task_done()

async def flow_parallel(*streams: AsyncIterator[float]) -> None:
    async def coro_fn(aitr: AsyncIterator[float]) -> None:
        async for s in aitr:
            await asyncio.sleep(s)
    awbls = (coro_fn(m) for m in streams)
    await asyncio.gather(*awbls)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable, Sequence, Tuple
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.flair import FlairTemplate, FlairTemplateChoices, UserFlairAssociation

import csv
from io import StringIO

from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.flair_async1 import UserFlairAssociationAsyncPaginator
from ...model_loaders.flair import (
    load_variant2_flair_template,
    load_variant1_flair_template,
    load_flair_template_choices,
    load_user_flair_association,
)

class FlairProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def set_user_flair(self,
        sr: str,
        user: str,
        text: Optional[str],
        css_class: Optional[str] = None,
    ) -> None:
        """Set the flair information on a user.

        Any previously set flair information will be discarded.

        A null argument means the field will not be sent in the request.
        The API treats this as the same as supplying an empty string.

        To revoke a flair, specify null or an empty string for the `text`
        and `css_class` parameters.

        .. .PARAMETERS

        :param `str` sr:
            Subreddit name.
        :param `str` user:
            User name.
        :param `Optional[str]` text:
            Flair text.

            Specify a string no longer than 64 characters.
            If longer than 64 characters then the parameter is ignored and an empty string is used.
        :param `Optional[str]` css_class:
            Flair CSS class.

            Specify a string no longer than 100 characters.
            If longer than 100 characters then a `BAD_CSS_NAME` API error is returned.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `BAD_FLAIR_TARGET`:
                The specified user doesn't exist.
            + `BAD_CSS_NAME`:
               - The specified CSS class was longer than 100 characters.
               - The specified CSS class contained invalid characters.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have permission to set flairs in the specified subreddit.

            + `404`:
                The specified subreddit does not exist.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('name', user)
            if text is not None: yield ('text', text)
            if css_class is not None: yield ('css_class', css_class)

        await self._client.request('POST', f'/r/{sr}/api/flair', data=dict(g()))

    async def set_post_flair(self,
        sr: str,
        subm: int,
        text: Optional[str],
        css_class: Optional[str] = None,
    ) -> None:
        """Set the flair information on a post.

        Any previously set flair information will be discarded.

        A null argument means the field will not be sent in the request.
        The API treats this as the same as supplying an empty string.

        To revoke a flair, specify null or an empty string for the `text`
        and `css_class` parameters.

        .. .PARAMETERS

        :param `str` sr:
            Subreddit name.
        :param `int` subm:
            Submission ID.
        :param `Optional[str]` text:
            Flair text.

            Specify a string no longer than 64 characters.
            If longer than 64 characters then the parameter is ignored and an empty string is used.
        :param `Optional[str]` css_class:
            Flair CSS class.

            Specify a string no longer than 100 characters.
            If longer than 100 characters then a `BAD_CSS_NAME` API error is returned.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `BAD_FLAIR_TARGET`:
                The specified submission doesn't exist.
            + `BAD_CSS_NAME`:
               - The specified CSS class was longer than 100 characters.
               - The specified CSS class contained invalid characters.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have permission to set flairs in the specified subreddit.
               - The target submission does not belong to the specified subreddit.

            + `404`:
                The specified subreddit does not exist.
        """
        full_id36 = 't3_' + to_base36(subm)

        def g() -> Iterable[tuple[str, str]]:
            yield ('link', full_id36)
            if text is not None: yield ('text', text)
            if css_class is not None: yield ('css_class', css_class)

        await self._client.request('POST', f'/r/{sr}/api/flair', data=dict(g()))

    def bulk_set_user_flairs(self,
        sr: str,
        items: Iterable[Tuple[str, str, str]],
    ) -> CallChunkChainingAsyncIterator[bool]:
        """Bulk set flair information for users.

        The second parameter is an iterable of 3-element tuples:
        `(user, text, css_class)`: the target user name, the flair text,
        and CSS class to assign.

        .. .PARAMETERS

        :param `str` sr:
        :param `Iterable[tuple[str, str, str]]` items:
            An iterable of tuples consisting of `(user, text, css_class)`.

        .. .RETURNS

        :returns:
            For each input item, a boolean appears in the output indicating whether
            the flair assignment succeeded.
        :rtype: :class:`~.iterators.call_chunk_chaining_async_iterator.CallChunkChainingAsyncIterator`\\[`bool`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have permission to set flairs in the specified subreddit.

            + `404`:
                The specified subreddit does not exist.
        """
        async def mass_update_user_flairs(items: Sequence[Tuple[str, str, str]]) -> Sequence[bool]:
            sio = StringIO()
            csv.writer(sio).writerows(items)
            s = sio.getvalue()
            root = await self._client.request('POST', f'/r/{sr}/api/flaircsv', files={'flair_csv': s})
            return [d['ok'] for d in root]

        itr = map(
            lambda xs: AsyncCallChunk(mass_update_user_flairs, xs),
            chunked(items, 100),
        )
        return CallChunkChainingAsyncIterator(itr)

    async def _create_or_update_flair_template(self,
        sr: str,
        flair_type: str,
        *,
        uuid: Optional[str] = None,
        text: Optional[str] = None,
        css_class: Optional[str] = None,
        bg_color: Optional[str] = None,
        fg_color_scheme: Optional[str] = None,
        mod_only: Optional[bool] = None,
        text_editable: Optional[bool] = None,
        allowable_content: Optional[str] = None,
        max_emojis: Optional[int] = None,
    ) -> FlairTemplate:
        d = {'flair_type': flair_type}
        for k, v in (
            ('flair_template_id', uuid),
            ('text', text),
            ('css_class', css_class),
            ('background_color', bg_color),
            ('text_color', fg_color_scheme),
            ('mod_only', None if mod_only is None else '01'[mod_only]),
            ('text_editable', None if text_editable is None else '01'[text_editable]),
            ('allowable_content', allowable_content),
            ('max_emojis', str(max_emojis)),
        ):
            if v is not None:
                d[k] = v
        root = await self._client.request('POST', f'/r/{sr}/api/flairtemplate_v2', data=d)
        return load_variant2_flair_template(root)

    async def create_user_flair_template(self,
        sr: str,
        *,
        text: Optional[str] = None,
        css_class: Optional[str] = None,
        bg_color: Optional[str] = None,
        fg_color_scheme: Optional[str] = None,
        mod_only: Optional[bool] = None,
        text_editable: Optional[bool] = None,
        allowable_content: Optional[str] = None,
        max_emojis: Optional[int] = None,
    ) -> FlairTemplate:
        """Create a user flair template.

        Any previously set flair information will be discarded.

        A null argument means the field will not be sent in the request.
        The API treats this as the same as supplying an empty string.

        .. .PARAMETERS

        :param `str` sr:
        :param `Optional[str]` text:
        :param `Optional[str]` css_class:
        :param `Optional[str]` bg_color:
            A 6-digit rgb hex color with an optional hash at the start. E.g. `#fb8559`.

            For user flair templates, the background color can be unset, making it transparent.
            (Post flairs cannot be transparent.)

            Effective default: empty string. It will be transparent.
        :param `Optional[str]` fg_color_scheme:
            Either `dark` or `light`.

            Effective default: `dark`.
        :param `Optional[bool]` mod_only:
            Whether flair is only available for mods to select.

            Effective default: false.
        :param `Optional[bool]` text_editable:
            Whether users will be able to edit their flair text.

            Effective default: false.
        :param `Optional[str]` allowable_content:
            Either: `all`, `emoji`, `text`.

            Effective default: `all`.
        :param `Optional[int]` max_emojis:
            An integer from 1 to 10.

            Effective default: `10`.

        .. .RETURNS

        :returns: The newly created flair template.
        :rtype: :class:`~.models.flair.FlairTemplate`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have permission to set flairs in the specified subreddit.

            + `404`:
                The specified subreddit does not exist.
        """
        return await self._create_or_update_flair_template(
            sr,
            flair_type='USER_FLAIR',
            text=text,
            css_class=css_class,
            bg_color=bg_color,
            fg_color_scheme=fg_color_scheme,
            mod_only=mod_only,
            text_editable=text_editable,
            allowable_content=allowable_content,
            max_emojis=max_emojis,
        )

    async def create_post_flair_template(self,
        sr: str,
        *,
        text: Optional[str] = None,
        css_class: Optional[str] = None,
        bg_color: Optional[str] = None,
        fg_color_scheme: Optional[str] = None,
        mod_only: Optional[bool] = None,
        text_editable: Optional[bool] = None,
        allowable_content: Optional[str] = None,
        max_emojis: Optional[int] = None,
    ) -> FlairTemplate:
        """Create a post flair template.

        Behaves similarly to :meth:`.create_user_flair_template`.

        .. .PARAMETERS

        :(parameters):
            Same as :meth:`.create_user_flair_template`,
            but `bg_color` effective default is different.

        :param `Optional[str]` bg_color:
            Effective default: `#d3d6da`.

        .. .RETURNS

        :(returns): Same as :meth:`.create_user_flair_template`.

        .. .RAISES

        :(raises): Same as :meth:`.create_user_flair_template`.
        """
        return await self._create_or_update_flair_template(
            sr,
            flair_type='LINK_FLAIR',
            text=text,
            css_class=css_class,
            bg_color=bg_color,
            fg_color_scheme=fg_color_scheme,
            mod_only=mod_only,
            text_editable=text_editable,
            allowable_content=allowable_content,
            max_emojis=max_emojis,
        )

    async def update_user_flair_template(self,
        sr: str,
        uuid: str,
        *,
        text: Optional[str] = None,
        css_class: Optional[str] = None,
        bg_color: Optional[str] = None,
        fg_color_scheme: Optional[str] = None,
        mod_only: Optional[bool] = None,
        text_editable: Optional[bool] = None,
        allowable_content: Optional[str] = None,
        max_emojis: Optional[int] = None,
    ) -> FlairTemplate:
        """Update a user flair template.

        Behaves similarly to :meth:`.create_user_flair_template`.

        .. .PARAMETERS

        :(parameters):
            Similar to :meth:`.create_user_flair_template`.

        :param `str` uuid:
            The flair template UUID.
        """
        return await self._create_or_update_flair_template(
            sr,
            flair_type='USER_FLAIR',
            uuid=uuid,
            text=text,
            css_class=css_class,
            bg_color=bg_color,
            fg_color_scheme=fg_color_scheme,
            mod_only=mod_only,
            text_editable=text_editable,
            allowable_content=allowable_content,
            max_emojis=max_emojis,
        )

    async def update_post_flair_template(self,
        sr: str,
        uuid: str,
        *,
        text: Optional[str] = None,
        css_class: Optional[str] = None,
        bg_color: Optional[str] = None,
        fg_color_scheme: Optional[str] = None,
        mod_only: Optional[bool] = None,
        text_editable: Optional[bool] = None,
        allowable_content: Optional[str] = None,
        max_emojis: Optional[int] = None,
    ) -> FlairTemplate:
        """Update a post flair template.

        Behaves similarly to :meth:`.update_user_flair_template`.
        """
        return await self._create_or_update_flair_template(
            sr,
            flair_type='LINK_FLAIR',
            uuid=uuid,
            text=text,
            css_class=css_class,
            bg_color=bg_color,
            fg_color_scheme=fg_color_scheme,
            mod_only=mod_only,
            text_editable=text_editable,
            allowable_content=allowable_content,
            max_emojis=max_emojis,
        )

    async def delete_flair_template(self, sr: str, uuid: str) -> None:
        """Delete a user or post flair template.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have permission.

            + `404`:
               - The specified subreddit does not exist.
               - The specified flair UUID does not exist.
        """
        await self._client.request('POST', f'/r/{sr}/api/deleteflairtemplate', data={'flair_template_id': uuid})

    async def delete_all_user_flair_templates(self, sr: str) -> None:
        """Delete all user flair templates in a subreddit.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have permission.

            + `404`:
                The specified subreddit does not exist.
        """
        await self._client.request('POST', f'/r/{sr}/api/clearflairtemplates', data={'flair_type': 'USER_FLAIR'})

    async def delete_all_post_flair_templates(self, sr: str) -> None:
        """Delete all post flair templates in a subreddit."""
        await self._client.request('POST', f'/r/{sr}/api/clearflairtemplates', data={'flair_type': 'LINK_FLAIR'})

    async def retrieve_user_flair_templates(self, sr: str) -> Sequence[FlairTemplate]:
        """Return a list of available user flair templates in a subreddit.

        Current user must be a moderator of the subreddit
        (otherwise a 403 HTTP error is returned).

        For non-mods, there is :meth:`.get_post_flair_choices`.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.flair.FlairTemplate`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
                The specified subreddit does not exist.
            + `403`:
                You do not have permission.
        """
        root = await self._client.request('GET', f'/r/{sr}/api/user_flair_v2')
        return [load_variant1_flair_template(d) for d in root]

    async def retrieve_post_flair_templates(self, sr: str) -> Sequence[FlairTemplate]:
        """Return a list of available post flair templates in a subreddit.

        Behaves similarly to :meth:`.retrieve_user_flair_templates`.
        """
        root = await self._client.request('GET', f'/r/{sr}/api/link_flair_v2')
        return [load_variant1_flair_template(d) for d in root]

    async def reorder_user_flair_templates(self, sr: str, order: Sequence[str]) -> None:
        """Reorder user flair templates.

        Reorders the flair templates as shown in the UI.

        The list must contain every flair UUID, otherwise a 400 HTTP error is returned.

        If you duplicate an ID the flair will have multiple references in the UI.

        .. .PARAMETERS

        :param `str` sr:
        :param `Sequence[str]` order:
            A list of all flair UUIDs.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                A flair template ID is missing from the provided list.
            + `500`:
                The specified subreddit does not exist.
        """
        params = {'subreddit': sr, 'flair_type': 'USER_FLAIR'}
        await self._client.request('PATCH', '/api/flair_template_order', params=params, json=order)

    async def reorder_post_flair_templates(self, sr: str, order: Sequence[str]) -> None:
        """Reorder post flair templates.

        Behaves similarly to :meth:`.reorder_user_flair_templates`.
        """
        params = {'subreddit': sr, 'flair_type': 'LINK_FLAIR'}
        await self._client.request('PATCH', '/api/flair_template_order', params=params, json=order)

    async def assign_user_flair_template(self,
        sr: str,
        user: str,
        uuid: str,
        *,
        text: Optional[str] = None,
    ) -> None:
        """Assign a user flair template.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `str` uuid:
        :param `Optional[str]` text:
            Custom text to override the template's text.

            Supply `None` or empty string to use the template's text.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The specified flair UUID does not exist.
               - You do not have permission.

            + `404`:
               - The specified subreddit does not exist.
               - The specified user does not exist.
        """
        d = {
            'name': user,
            'flair_template_id': uuid,
        }
        if text is not None:
            d['text'] = text
        await self._client.request('POST', f'/r/{sr}/api/selectflair', data=d)

    async def assign_post_flair_template(self,
        sr: str,
        subm: int,
        uuid: str,
        *,
        text: Optional[str] = None,
    ) -> None:
        """Assign a post flair template.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `str` uuid:
        :param `Optional[str]` text:
            Custom text to override the template's text.

            Supply `None` or empty string to use the template's text.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The specified flair UUID does not exist.
               - You do not have permission.

            + `404`:
               - The specified subreddit does not exist.
               - The specified submission does not exist.
        """
        full_id36 = 't3_' + to_base36(subm)
        d = {
            'link': full_id36,
            'flair_template_id': uuid,
        }
        if text is not None:
            d['text'] = text
        await self._client.request('POST', f'/r/{sr}/api/selectflair', data=d)

    async def assign_user_flair(self,
        sr: str,
        user: str,
        text: Optional[str],
        css_class: Optional[str] = None,
        *,
        bg_color: Optional[str] = None,
        fg_color_scheme: Optional[str] = None,
    ) -> None:
        """Assign a custom user flair.

        This is a newer version of :meth:`.set_user_flair`.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `Optional[str]` text:
        :param `Optional[str]` css_class:
        :param `Optional[str]` bg_color:
        :param `Optional[str]` fg_color_scheme:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `BAD_CSS_NAME`:
               - The specified CSS class was longer than 100 characters.
               - The specified CSS class contained invalid characters.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `404`:
               - The specified subreddit does not exist.
               - The specified user does not exist.
        """
        d = {'name': user}
        for k, v in (
            ('text', text),
            ('css_class', css_class),
            ('background_color', bg_color),
            ('text_color', fg_color_scheme),
        ):
            if v is not None:
                d[k] = v
        await self._client.request('POST', f'/r/{sr}/api/selectflair', data=d)

    async def assign_post_flair(self,
        sr: str,
        subm: int,
        text: Optional[str],
        css_class: Optional[str] = None,
        *,
        bg_color: Optional[str] = None,
        fg_color_scheme: Optional[str] = None,
    ) -> None:
        """Assign a custom post flair.

        This is a newer version of :meth:`.set_user_flair`.

        .. .PARAMETERS

        :param `str` sr:
        :param `int` subm:
        :param `Optional[str]` text:
        :param `Optional[str]` css_class:
        :param `Optional[str]` bg_color:
        :param `Optional[str]` fg_color_scheme:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `BAD_CSS_NAME`:
               - The specified CSS class was longer than 100 characters.
               - The specified CSS class contained invalid characters.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `404`:
               - The specified subreddit does not exist.
               - The specified submission does not exist.
        """
        full_id36 = 't3_' + to_base36(subm)
        d = {'link': full_id36}
        for k, v in (
            ('text', text),
            ('css_class', css_class),
            ('background_color', bg_color),
            ('text_color', fg_color_scheme),
        ):
            if v is not None:
                d[k] = v
        await self._client.request('POST', f'/r/{sr}/api/selectflair', data=d)

    async def configure_subreddit_flair_settings(self,
        sr: str,
        *,
        user_enabled: Optional[bool] = False,
        user_position: Optional[str] = '',
        user_self_assign: Optional[bool] = False,
        post_position: Optional[str] = '',
        post_self_assign: Optional[bool] = False,
    ) -> None:
        """Configure subreddit flair settings.

        All parameters should be specified. If a parameter is not specified or is
        an invalid value, its default will be used.

        User flairs are disabled when either `user_enabled` is false
        or `user_position` is an empty string.

        Post flairs are disabled when `post_position` is an empty string.

        .. .PARAMETERS

        :param `str` sr:
        :param `Optional[bool]` user_enabled:
            Whether user flairs are enabled in the subreddit.

            Effective default: false.
        :param `Optional[str]` user_position:
            Either `left`, `right`, or empty string.

            An empty string value will disable user flairs even if
            `user_enabled` is true.

            Effective default: empty string.
        :param `Optional[bool]` user_self_assign:
            Whether users are allowed to assign their own user flairs.

            Value is forced false if `user_enabled` is false
            (but not if `user_position` is an empty string).

            Effective default: false.
        :param `Optional[str]` post_position:
            Either `left`, `right`, or empty string.

            Effective default: empty string.
        :param `Optional[bool]` post_self_assign:
            Whether users are allowed to assign their own post flairs.

            Value is forced false if `post_position` is an empty string.

            Effective default: false.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `404`:
                The specified subreddit does not exist.
        """
        def g() -> Iterable[tuple[str, str]]:
            if user_enabled is not None:
                yield ('flair_enabled', '01'[user_enabled])
            if user_position is not None:
                yield ('flair_position', user_position)
            if user_self_assign is not None:
                yield ('flair_self_assign_enabled', '01'[user_self_assign])
            if post_position is not None:
                yield ('link_flair_position', post_position)
            if post_self_assign is not None:
                yield ('link_flair_self_assign_enabled', '01'[post_self_assign])

        await self._client.request('POST', f'/r/{sr}/api/flairconfig', data=dict(g()))

    async def get_user_flair_template_choices(self, sr: str) -> FlairTemplateChoices:
        """Get user flair template choices.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :returns: :class:`~.models.flair.FlairTemplateChoices`, a `Sequence`\\[:class:`~.models.flair.FlairTemplateChoice`]
        :rtype: :class:`~.models.flair.FlairTemplateChoices`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
                There is no user context.
            + `404`:
                The specified subreddit does not exist.
        """
        root = await self._client.request('POST', f'/r/{sr}/api/flairselector')
        return load_flair_template_choices(root)

    async def get_post_flair_template_choices(self, sr: str) -> FlairTemplateChoices:
        """Get user flair template choices.

        Behaves similarly to :meth:`.get_user_flair_template_choices`.
        """
        root = await self._client.request('POST', f'/r/{sr}/api/flairselector', data={'is_newlink': '1'})
        return load_flair_template_choices(root)

    async def get_user_flair_association(self, sr: str, user: str) -> Optional[UserFlairAssociation]:
        """Get a user flair association in a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.flair.UserFlairAssociation`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
                The specified subreddit does not exist.
            + `403`:
               - There is no user context.
               - You are not a moderator of the subreddit.
        """
        params = {'name': user, 'limit': '1'}
        root = await self._client.request('GET', f'/r/{sr}/api/flairlist', params=params)
        users_data = root['users']
        if not users_data:
            return None
        user_data = users_data[0]
        if user_data['user'].lower() != user.lower():
            return None
        return load_user_flair_association(user_data)

    def get_user_flair_associations(self,
        sr: str,
        amount: Optional[int] = None,
    ) -> ImpartedPaginatorChainingAsyncIterator[UserFlairAssociationAsyncPaginator, UserFlairAssociation]:
        """Get a user flair associations in a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_async_iterator.ImpartedPaginatorChainingAsyncIterator`\\[:class:`~.pagination.paginators.flair_async1.UserFlairAssociationAsyncPaginator`, :class:`~.models.flair.UserFlairAssociation`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
                The specified subreddit does not exist.
            + `403`:
               - There is no user context.
               - You are not a moderator of the subreddit.
        """
        p = UserFlairAssociationAsyncPaginator(self._client, f'/r/{sr}/api/flairlist')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    async def show_my_flair(self, sr: str) -> None:
        """Show the current user's flair in the subreddit.

        To tell if the current user's flair is shown::

           subr = await client.p.subreddit.fetch_by_name('Python')
           print(subr.me.flair.shown)

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `404`:
                The specified subreddit does not exist.
        """
        await self._client.request('GET', f'/r/{sr}/api/setflairenabled', params={'flair_enabled': '1'})

    async def hide_my_flair(self, sr: str) -> None:
        """Hide the current user's flair in the subreddit."""
        await self._client.request('GET', f'/r/{sr}/api/setflairenabled', params={'flair_enabled': '0'})




from __future__ import annotations
from typing import Optional

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class ImageInfo:
    url: str
    width: Optional[int]
    height: Optional[int]
    name: str




from __future__ import annotations
from typing import Mapping, Any, Optional

from datetime import datetime, timezone

from .datamemento import DatamementoPropertiesMixin

class ModeratedSubreddit(DatamementoPropertiesMixin):
    class Me:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.is_subscribed: bool = d['user_is_subscriber']
            ("")

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        full_id36: str = d['name']
        self.id36: str = full_id36[3:]
        ("""
            The ID of the subreddit as a base 36 number.
            """)
        self.id: int = int(self.id36, 36)
        ("""
            The subreddit ID.
            """)
        self.name: str = d['display_name']
        ("""
            The name of the subreddit.
            """)
        self.openness: str = d['subreddit_type']
        ("""
            Either: `public`, `private`, `restricted`, `archived`,
            `employees_only`, `gold_only`, `gold_restricted`, or `user`.
            """)
        self.created_ut: int = -1
        ("""
            If a normal subreddit, the unix timestamp of when the subreddit was created.

            If a user subreddit, the value is `-1`.

            Use `self.openness == 'user'` to tell if a subreddit is a user subreddit.
            """)
        self.created_at: datetime = datetime.min
        ("""
            If a normal subreddit, a datetime object of when the subreddit was created.

            If a user subreddit, the value is `datetime.min`.

            Use `self.openness == 'user'` to tell if a subreddit is a user subreddit.
            """)
        if self.openness != 'user':
            self.created_ut = int(d['created_utc'])
            self.created_at = datetime.fromtimestamp(self.created_ut, timezone.utc)

        self.title: str = d['title']
        ("")
        self.subscriber_count: int = d['subscribers']
        ("")
        self.nsfw: bool = d['over_18']
        ("")
        self.icon_img: str = d['icon_img']
        ("")

        self.me: Optional[ModeratedSubreddit.Me] = None
        ("")
        if 'user_is_subscriber' in d:
            self.me = self.Me(d)




from typing import Mapping, Any

from ..models.media_upload_lease import MediaUploadLease

def load_media_upload_lease(d: Mapping[str, Any]) -> MediaUploadLease:
    lease_data = d['args']
    endpoint = f'https:{x}' if (x := lease_data['action']).startswith('//') else x
    fields = {field['name']: field['value'] for field in lease_data['fields']}
    s3_object_key = fields['key']
    asset = d['asset']
    return MediaUploadLease(
        d=d,
        endpoint=endpoint,
        fields=fields,
        s3_object_key=fields['key'],
        location=f"{endpoint}/{s3_object_key}",
        media_id=asset['asset_id'],
        declared_filepath=asset['payload']['filepath'],
        websocket_url=asset['websocket_url'],
    )



#!/usr/bin/env python
from __future__ import annotations
from typing import TYPE_CHECKING, MutableSequence, Iterator, Callable
if TYPE_CHECKING:
    from redditwarp.models.comment_tree_SYNC import CommentSubtreeTreeNode, MoreCommentsTreeNode

###
ALGORITHM_CHOICES = (
    'dfs',
    'dfs0',
    'dfs1',
    'dfs2',
    'bfs',
    'bfs1',
    'bfs2',
)
COMMENT_SORT_CHOICES = 'confidence top new controversial old random qa live'.split()

import argparse
class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter): pass
parser = argparse.ArgumentParser(description=__doc__, formatter_class=Formatter)
parser._optionals.title = __import__('gettext').gettext('named arguments')
add = parser.add_argument
add('target', nargs='?')
add('--access-token', metavar='ACCESS_TOKEN', dest='access_token_opt')
add('--target', metavar='TARGET', dest='target_opt', help=argparse.SUPPRESS)
add('--base', metavar='N', dest='base_opt', type=int, default=36)
add('--algo', '--algorithm', choices=ALGORITHM_CHOICES, default='dfs', dest='algo', metavar='')
add('--sort', dest='sort', choices=COMMENT_SORT_CHOICES, default='confidence')
args = parser.parse_args()
###;

import sys
import shutil
from collections import deque

import redditwarp.SYNC
from redditwarp.util.extract_id_from_url import extract_submission_id_from_url
from redditwarp.models.comment_SYNC import Comment

access_token: str = args.access_token_opt or ''
target: str = args.target or args.target_opt or input('Target: ')
algo: str = args.algo
sort: str = args.sort
base: int = args.base_opt


# A recursive DFS. The algorithm looks clean and performs well, except that Python
# limits the maximum function recursion depth to 1000, so if you have a very deep
# comment tree you'll have to use an iterative approach.
def depth_first_recursive(node: CommentSubtreeTreeNode[object]) -> Iterator[tuple[int, Comment]]:
    def traverse(root: CommentSubtreeTreeNode[object], level: int = 0) -> Iterator[tuple[int, Comment]]:
        value = root.value
        if isinstance(value, Comment):
            yield (level, value)

        for child in root.children:
            yield from traverse(child, level + 1)

        if root.more:
            yield from traverse(root.more(), level)

    return traverse(node)

# An iterative DFS that is functionally equivalent to the recursive version but is
# slightly inaccurate because the `MoreComments` callables are evaluated before the
# child nodes are processed. It's undesirable because for a display script it has the
# effect of feeling slower and looking more jittery.
def depth_first_iterative_inaccurate(node: CommentSubtreeTreeNode[object]) -> Iterator[tuple[int, Comment]]:
    stack: MutableSequence[CommentSubtreeTreeNode[object]] = deque([node])
    levels = deque([0])
    while stack:
        node = stack.pop()
        level = levels.pop()

        value = node.value
        if isinstance(value, Comment):
            yield (level, value)

        if node.more:
            stack.append(node.more())
            levels.append(level)

        stack.extend(reversed(node.children))
        levels.extend([level + 1] * len(node.children))

# This version is more algorithmically accurate to the recursive one but at the cost of
# looking messier.
def depth_first_iterative_accurate(node: CommentSubtreeTreeNode[object]) -> Iterator[tuple[int, Comment]]:
    stack: MutableSequence[bool] = deque([True])
    node_stack: MutableSequence[CommentSubtreeTreeNode[object]] = deque([node])
    more_stack: MutableSequence[Callable[[], MoreCommentsTreeNode]] = deque()
    levels = deque([0])
    while stack:
        if stack.pop():
            node = node_stack.pop()
        else:
            node = more_stack.pop()()
        level = levels.pop()

        value = node.value
        if isinstance(value, Comment):
            yield (level, value)

        if node.more:
            stack.append(False)
            more_stack.append(node.more)
            levels.append(level)

        stack.extend([True] * len(node.children))
        node_stack.extend(reversed(node.children))
        levels.extend([level + 1] * len(node.children))

# This BFS traversal evaluates the `MoreComments` callables before processing the
# children which, we've established is undesirable because it feels slow.
def breadth_first_inaccurate(node: CommentSubtreeTreeNode[object]) -> Iterator[tuple[int, Comment]]:
    level = 0
    queue: MutableSequence[CommentSubtreeTreeNode[object]] = deque([node])
    while queue:
        batch = deque(queue)
        queue.clear()
        while batch:
            node = batch.popleft()
            if node.value is None:
                if node.more:
                    batch.appendleft(node.more())
                batch.extendleft(reversed(node.children))
                continue

            value = node.value
            if isinstance(value, Comment):
                yield (level, value)

            queue.extend(node.children)
            if node.more:
                queue.append(node.more())

        level += 1

# A BFS that processes all children before evaluating `MoreComments` callables.
# Algorithmically better but programmatically uglier.
def breadth_first_accurate(node: CommentSubtreeTreeNode[object]) -> Iterator[tuple[int, Comment]]:
    level = 0
    queue: MutableSequence[bool] = deque([True])
    node_queue: MutableSequence[CommentSubtreeTreeNode[object]] = deque([node])
    more_queue: MutableSequence[Callable[[], MoreCommentsTreeNode]] = deque()
    while queue:
        batch = deque(queue)
        node_batch = deque(node_queue)
        more_batch = deque(more_queue)
        queue.clear()
        node_queue.clear()
        more_queue.clear()
        while batch:
            if batch.popleft():
                node = node_batch.pop()
            else:
                node = more_batch.pop()()

            if node.value is None:
                if node.more:
                    batch.appendleft(False)
                    more_batch.appendleft(node.more)
                batch.extendleft([True] * len(node.children))
                node_batch.extendleft(reversed(node.children))
                continue

            value = node.value
            if isinstance(value, Comment):
                yield (level, value)

            queue.extend([True] * len(node.children))
            node_queue.extend(node.children)
            if node.more:
                queue.append(False)
                more_queue.append(node.more)

        level += 1


traversal = {
    'dfs': depth_first_iterative_accurate,
    'dfs0': depth_first_recursive,
    'dfs1': depth_first_iterative_inaccurate,
    'dfs2': depth_first_iterative_accurate,
    'bfs': breadth_first_accurate,
    'bfs1': breadth_first_inaccurate,
    'bfs2': breadth_first_accurate,
}[algo]

client = (
    redditwarp.SYNC.Client.from_access_token(access_token)
    if access_token else
    redditwarp.SYNC.Client()
)

client.http.set_user_agent(client.http.get_user_agent() + " redditwarp.cli.comment_tree")

idn = 0
if target.isalnum():
    try:
        idn = int(target, base)
    except ValueError as e:
        print(f'{e.__class__.__name__}: {e}', file=sys.stderr)
        sys.exit(1)
else:
    try:
        idn = extract_submission_id_from_url(target)
    except ValueError:
        print('Could not extract comment ID from URL.', file=sys.stderr)
        sys.exit(1)

tree_node = client.p.comment_tree.get(idn, sort=sort)
if tree_node is None:
    print('Submission not found', file=sys.stderr)
    sys.exit(1)

m = tree_node.value

print(f'''\
{m.permalink}
{m.id36}+ ^{m.score} | {m.title}
Submitted {m.created_at.astimezone().ctime()}{' *' if m.is_edited else ''} \
by u/{m.author_display_name} to r/{m.subreddit.name}
''')

columns, _lines = shutil.get_terminal_size()

for depth, comment in traversal(tree_node):
    c = comment
    body_text = repr(c.body)
    line = f"{depth*'.'} u/{c.author_display_name} | {body_text}"
    print(line[:columns])




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from collections.abc import Iterable, Collection, Sequence, Set, Mapping
    from ..client_ASYNC import Client

from ...pagination.async_paginator import AsyncPaginator
from ..models.document import Document


class DocumentAsyncPaginator(AsyncPaginator[Document]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 1000,
        search_params: Optional[Mapping[str, str]] = None,
        since: Optional[int] = None,
        until: Optional[int] = None,
        fields: Optional[Collection[str]] = None,
        descending: bool = False,
    ) -> None:
        super().__init__(limit=limit)
        self.client: Client = client
        ("")
        self.url: str = url
        ("")
        self._search_params: Mapping[str, str] = {} if search_params is None else search_params
        self._since_cursor: Optional[int] = since
        self._until_cursor: Optional[int] = until
        self._fields: Optional[Collection[str]] = fields
        self._descending: bool = descending
        self._previous_document_ids: Set[int] = set()

    async def fetch(self) -> Sequence[Document]:
        def g() -> Iterable[tuple[str, str]]:
            yield ('order', 'desc' if self._descending else 'asc')
            if self.limit is not None: yield ('limit', str(self.limit))
            if self._since_cursor is not None: yield ('since', str(self._since_cursor))
            if self._until_cursor is not None: yield ('until', str(self._until_cursor))
            if self._fields: yield ('filter', ','.join(self._fields))
            yield from self._search_params.items()

        root = await self.client.request('GET', self.url, params=dict(g()))
        documents = root['data']

        m = {d['id']: d for d in documents}
        for i in self._previous_document_ids:
            m.pop(i, None)
        self._previous_document_ids = set(m)
        documents = list(m.values())

        # If this is true then we could be losing some results unless we
        # increase the fetch limit. But the pagination algorithm shouldn't
        # alter the limit so just carry on and try our best.
        all_from_same_time_bucket: bool = len(set(d['created_utc'] for d in documents)) == 1

        if documents:
            cursor: int = documents[-1]['created_utc']
            if self._descending:
                if (self._until_cursor is None) or cursor < self._until_cursor:
                    self._until_cursor = cursor + 1
                if all_from_same_time_bucket:
                    self._until_cursor -= 1
            else:
                if (self._since_cursor is None) or cursor > self._since_cursor:
                    self._since_cursor = cursor
                if all_from_same_time_bucket:
                    self._since_cursor += 1

        return [Document(d) for d in documents]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterator, overload, Union
if TYPE_CHECKING:
    from .menu_bar_SYNC import MenuBar
    from .community_details_SYNC import CommunityDetailsWidget
    from .moderator_list_SYNC import ModeratorListWidget
    from .rules_SYNC import RulesWidget

from .widget_SYNC import Widget

class WidgetList(Sequence[Widget]):
    def __init__(self,
        *,
        widgets: Sequence[Widget],
        menu_bar: Optional[MenuBar],
        community_details_widget: CommunityDetailsWidget,
        moderator_list_widget: ModeratorListWidget,
        rules_widget: Optional[RulesWidget],
    ) -> None:
        self._widgets = widgets
        self.menu_bar: Optional[MenuBar] = menu_bar
        self._community_details_widget = community_details_widget
        self._moderator_list_widget = moderator_list_widget
        self._rules_widget = rules_widget

    def __len__(self) -> int:
        return len(self._widgets)

    def __contains__(self, item: object) -> bool:
        return item in self._widgets

    def __iter__(self) -> Iterator[Widget]:
        return iter(self._widgets)

    @overload
    def __getitem__(self, index: int) -> Widget: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Widget]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Widget, Sequence[Widget]]:
        return self._widgets[index]

    def get_community_details_widget(self) -> CommunityDetailsWidget:
        return self._community_details_widget

    def get_moderator_list_widget(self) -> ModeratorListWidget:
        return self._moderator_list_widget

    def get_rules_widget(self) -> Optional[RulesWidget]:
        return self._rules_widget




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.modmail_ASYNC import ConversationInfo, Message
    from ..stream_ASYNC import IStandardStreamEventSubject
    from ...pagination.async_paginator import CursorAsyncPaginator

from ..stream_ASYNC import Stream


def conversation_message_new_extractor(x: tuple[ConversationInfo, Message]) -> object:
    return (x[0].id, x[1].id)

def get_conversation_message_new_stream_paginator(client: Client) -> CursorAsyncPaginator[tuple[ConversationInfo, Message]]:
    return client.p.modmail.pull.new().get_paginator()

def make_conversation_message_new_stream(paginator: CursorAsyncPaginator[tuple[ConversationInfo, Message]], past: Optional[Iterable[tuple[ConversationInfo, Message]]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[tuple[ConversationInfo, Message]]:
    return Stream(paginator, conversation_message_new_extractor, past=past, seen=seen)

def create_conversation_message_new_stream(client: Client) -> IStandardStreamEventSubject[tuple[ConversationInfo, Message]]:
    return make_conversation_message_new_stream(get_conversation_message_new_stream_paginator(client))


def conversation_message_join_request_extractor(x: tuple[ConversationInfo, Message]) -> object:
    return (x[0].id, x[1].id)

def get_conversation_message_join_request_stream_paginator(client: Client) -> CursorAsyncPaginator[tuple[ConversationInfo, Message]]:
    return client.p.modmail.pull.join_requests().get_paginator()

def make_conversation_message_join_request_stream(paginator: CursorAsyncPaginator[tuple[ConversationInfo, Message]], past: Optional[Iterable[tuple[ConversationInfo, Message]]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[tuple[ConversationInfo, Message]]:
    return Stream(paginator, conversation_message_join_request_extractor, past=past, seen=seen)

def create_conversation_message_join_request_stream(client: Client) -> IStandardStreamEventSubject[tuple[ConversationInfo, Message]]:
    return make_conversation_message_join_request_stream(get_conversation_message_join_request_stream_paginator(client))




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.moderation_note import (
        ModerationNote,
        ModerationUserNote,
    )

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.moderation.note_pulls_sync1 import (
    ModerationNotePaginator,
    ModerationUserNotePaginator,
)
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_iterator import CallChunkChainingIterator
from ...iterators.call_chunk import CallChunk
from ...model_loaders.moderation_note import load_moderation_user_note
from ... import http

class Note:
    def __init__(self, client: Client) -> None:
        self._client = client

    def create_user_note(self, sr: str, user: str, content: str, *,
            label: str = '',
            anchor_submission_id: Optional[int] = None,
            anchor_comment_id: Optional[int] = None) -> ModerationUserNote:
        """Create a user note.

        .. .PARAMETERS

        :param `str` sr:
            Subreddit name.
        :param `str` user:
            User name.
        :param `str` content:
            Content of the note. Max 250 characters.
        :param `str` label:
            Either: `BOT_BAN`, `PERMA_BAN`, `BAN`, `ABUSE_WARNING`, `SPAM_WARNING`,
            `SPAM_WATCH`, `SOLID_CONTRIBUTOR`, `HELPFUL_USER`.
        :param `Optional[str]` anchor_submission_id:
            A submission ID to link the note to.

            Mutually exclusive with `anchor_comment_id`.
        :param `Optional[str]` anchor_comment_id:
            A comment ID to link the note to.

            Mutually exclusive with `anchor_submission_id`.

        .. .RETURNS

        :rtype: :class:`~.models.moderation_note.ModerationUserNote`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
            + `BAD_SR_NAME`:
                The specified subreddit does not exist.
            + `NO_TEXT`:
                The content of the note was empty.
            + `TOO_LONG`:
                The content of the note was too long, over 250 characters.
            + `INVALID_OPTION`:
                The label specified was invalid.
            + `NO_THING_ID`:
                The linked submission or comment does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You are not a moderator of the target subreddit.
        """
        mxp = (anchor_submission_id, anchor_comment_id)
        if len(mxp) - mxp.count(None) > 1:
            raise TypeError("mutually exclusive parameters: `anchor_submission_id`, `anchor_comment_id`.")

        params = {'subreddit': sr, 'user': user, 'note': content, 'label': label}
        root = self._client.request('POST', '/api/mod/notes', params=params)
        return load_moderation_user_note(root['created'])

    def delete(self, sr: str, user: str, uuid: str) -> None:
        """Delete a moderation note.

        This procedure can be used to delete either a user or action type note.

        In addition to the note ID, the endpoint must be given the subreddit and
        user name in which the note belongs. If either information is incorrect,
        the endpoint will raise a 500 HTTP error.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `str` uuid:
            UUID of a note.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `BAD_SR_NAME`:
                The `sr` parameter was empty.
            + `SUBREDDIT_NOEXIST`:
                The target subreddit does not exist.
            + `USER_DOESNT_EXIST`:
               - The specified user does not exist.
               - The `user` parameter was empty.
            + `PARAMETER_REQUIRED`:
                The `uuid` parameter was empty.
            + `INVALID_ID`:
                The specified user does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `500`:
               - The subreddit or user specified is incorrect for the given note ID.
               - The given note ID was invalid.
        """
        params = {'subreddit': sr, 'user': user, 'note_id': 'ModNote_' + uuid}
        self._client.request('DELETE', '/api/mod/notes', params=params)

    def pull_notes(self,
        sr: str,
        user: str,
        amount: Optional[int] = None,
        *,
        label: Optional[str] = None,
    ) -> ImpartedPaginatorChainingIterator[ModerationNotePaginator, ModerationNote]:
        """Get moderation notes of a user in a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:
        :param `Optional[int]` amount:
        :param `Optional[str]` label:
            Filter by note type.

            Either: `ALL`, `NOTE`, `APPROVAL`, `REMOVAL`, `BAN`, `MUTE`,
            `INVITE`, `SPAM`, `CONTENT_CHANGE`.

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`\\[:class:`~.pagination.paginators.moderation.note_pulls_sync1.ModerationNotePaginator`, :class:`~.models.moderation_note.ModerationNote`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `BAD_SR_NAME`:
                The `sr` parameter was empty.
            + `SUBREDDIT_NOEXIST`:
                The target subreddit does not exist.
            + `USER_DOESNT_EXIST`:
               - The specified user does not exist.
               - The `user` parameter was empty.
            + `INVALID_ID`:
                An invalid value was specified for the `label` parameter.
                Values are case-sensitive.
        """
        p = ModerationNotePaginator(
            client=self._client,
            url='/api/mod/notes',
            subreddit=sr,
            user=user,
            type=label,
        )
        return ImpartedPaginatorChainingIterator(p, amount)

    def pull_user_notes(self,
        sr: str,
        user: str,
        amount: Optional[int] = None,
    ) -> ImpartedPaginatorChainingIterator[ModerationUserNotePaginator, ModerationUserNote]:
        """
        Behaves similarly to :meth:`.pull_notes`.
        """
        p = ModerationUserNotePaginator(
            client=self._client,
            url='/api/mod/notes',
            subreddit=sr,
            user=user,
            type='NOTE',
        )
        return ImpartedPaginatorChainingIterator(p, amount)

    def get_latest_user_note(self, sr: str, user: str) -> Optional[ModerationUserNote]:
        """Get the most recently written user note for some user.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` user:

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.moderation_note.ModerationUserNote`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `PARAMETER_REQUIRED`:
               - The `sr` parameter was empty.
               - The `user` parameter was empty.
        """
        params = {'subreddits': sr, 'users': user}
        try:
            root = self._client.request('GET', '/api/mod/notes/recent', params=params)
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 400:
                return None
            raise
        note_obj = root['mod_notes'][0]
        if note_obj is None:
            return None
        return load_moderation_user_note(note_obj)

    def bulk_get_latest_user_note(self, pairs: Iterable[tuple[str, str]]) -> CallChunkChainingIterator[Optional[ModerationUserNote]]:
        """Bulk fetch the most recently written user notes for users.

        The response contains a list of mod notes in the order that subreddits and users were given.
        If no note exists for a given pair, `None` will take its place in the list.

        .. .PARAMETERS

        :param `Iterable[tuple[str, str]]` pairs:

        .. .RETURNS

        :rtype: :class:`~.iterators.call_chunk_chaining_iterator.CallChunkChainingIterator`\\[`Optional`\\[:class:`~.models.moderation_note.ModerationUserNote`]]

        .. .RAISES

        :(raises): Same as :meth:`.get_latest_user_note`.
        """
        def mass_get_latest_user_note(pairs: Sequence[tuple[str, str]]) -> Sequence[Optional[ModerationUserNote]]:
            subreddits, users = tuple(zip(*pairs))
            subreddits_str = ','.join(subreddits)
            users_str = ','.join(users)
            params = {'subreddits': subreddits_str, 'users': users_str}
            try:
                root = self._client.request('GET', '/api/mod/notes/recent', params=params)
            except http.exceptions.StatusCodeException as e:
                if e.status_code == 400:
                    return [None]
                raise
            return [
                (None if note_obj is None else load_moderation_user_note(note_obj))
                for note_obj in root['mod_notes']
            ]

        return CallChunkChainingIterator(CallChunk(mass_get_latest_user_note, chunk) for chunk in chunked(pairs, 500))




from __future__ import annotations
from typing import Mapping, Any

from ..models.flair_emoji import FlairEmojiUploadLease, FlairEmoji

def load_flair_emoji_upload_lease(d: Mapping[str, Any]) -> FlairEmojiUploadLease:
    lease_data = d['s3UploadLease']
    endpoint = f'https:{x}' if (x := lease_data['action']).startswith('//') else x
    fields = {field['name']: field['value'] for field in lease_data['fields']}
    s3_object_key = fields['key']
    return FlairEmojiUploadLease(
        d=d,
        endpoint=endpoint,
        fields=fields,
        s3_object_key=s3_object_key,
        location=f"{endpoint}/{s3_object_key}",
    )

def load_flair_emoji(d: Mapping[str, Any], name: str) -> FlairEmoji:
    full_id36: str = d['created_by']
    _, _, id36 = full_id36.partition('_')
    return FlairEmoji(
        d=d,
        name=name,
        image_url=d['url'],
        creator_id=int(id36, 36),
        creator_id36=id36,
        post_enabled=d['post_flair_allowed'],
        user_enabled=d['user_flair_allowed'],
        mod_only=d['mod_flair_only'],
    )
